/************************************************************************
 *
 *                       u_code.c Version 1.8
 *
 * NOTICE: Copyright (C) 1996 to 2001 by Robert S. Larkin
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * The program author can be contacted by email at boblark@proaxis.com.
 * Alternatively contact the author at
 *     Robert S. (Bob) Larkin
 *     2982 N. W. Acacia Place
 *     Corvallis, OR  97330  USA
 *
 * For more details on the background or use of the program, pay a visit to
 * http://www.proaxis.com/~boblark/dsp10.htm  and,
 * http://www.ao.com/~gnome/
 *
 *
 * Routines for sending and receiving codes with DSP10 transceiver
 * Presently supports:
 *     1 - Morse Code
 *     2 - LHL Code/modulation
 *     3 - PUA43 Code/modulation
 *     4 - EME2 2 sec on, 3 sec off, self echos
 *
 * Bob Larkin
 * 12 Sept 98   Original write of LHL & PUA43
 * 05 Dec  98   Revised to integrate send and rcv fcns in one module
 * 14 Sept 99   Temporarily pulled PUA & LHL code transmission.
 * 23 Jan  00   Added KN, Put PUA/LHL back.
 * 18 Feb  00   Added EME2
 * 23 Mar  00   Redid LHL7 for multiple bins, boxes, peak/ave and 2 line.
 * 2  Jan  01   Added LTI as a separate mode, no longer USB sub-mode.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <dos.h>
#include <graphics.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <bios.h>
#include "moonsun.h"
#include "async.h"
#include "uhfa.h"
#include "util.h"
#include "u_code.h"
#include "lbox.h"

void morse_ch(unsigned char,		/* Character to be sent */
	      int);			/* Code speed index (0, 25) */
void morse(int, int);			/* Code routines, (command, ichar) */

int  lhl_code(unsigned char, struct lhl*); /* Transmit LHL Code */
double pua43_freq(unsigned char);	/* Transmit PUA43 Code */
void lhl7_update(int, int);		/* Receive LHL Code */
void pua43_update(int, int);		/* Receive PUA43 Code */
void eme2_update(int);			/* Receive eme 2 sec transmissions */
void lterm_int_update(int);		/* Long term integration in USB */
void eme2_modalbox(void);
void lhl7_modalbox(void);
void pua43_modalbox(void);
void lterm_int_modalbox(void);

static void send(int cw);		/* Hdw control, cw=1 --> key down */

/******************   Global Variables ******************/
double pua43_colorscale=1.2;

struct lbox_info   lb1, lb2, lb3, lb4;

/* Note:  Code speed can be any.  There are 4 times in milliseconds that
 * set the actual speed and weighting.  The index (0, 25) allows 26 of these
 * that are initialized to 5 WPM to 30 WPM with nominal weightings.
 *
 * The following set the weightings vs. code speed in wpm
 * by selecting the delays for
 * t.dot   length of a dot in ms
 * t.dash  length of a dash in ms
 * t.dd    time between dots and dashes, ms
 * t.ch    time between characters of the same word, ms
 * t.sp    time duration between words, ms
 */
struct cw_data	t[26] = {
		 "5 WPM", 150, 530, 150, 900,1200,
		 "6 WPM", 140, 480, 140, 750,1000,
		 "7 WPM", 125, 430, 125, 550, 800,
		 "8 WPM", 120, 400, 120, 400, 725,
		 "9 WPM", 120, 380, 120, 300, 600,
		"10 WPM", 120, 360, 120, 240, 480,
		"11 WPM", 110, 330, 110, 220, 440,
		"12 WPM", 100, 200, 100, 200, 400,
		"13 WPM",  92, 276,  92, 184, 368,
		"14 WPM",  86, 258,  86, 172, 344,
		"15 WPM",  80, 240,  80, 160, 320,
		"16 WPM",  75, 225,  75, 150, 300,
		"17 WPM",  70, 212,  70, 142, 283,
		"18 WPM",  66, 200,  66, 134, 267,
		"19 WPM",  63, 189,  63, 126, 252,
		"20 WPM",  60, 180,  60, 120, 240,
		"21 WPM",  57, 171,  57, 114, 230,
		"22 WPM",  55, 164,  55, 109, 220,
		"23 WPM",  52, 157,  52, 105, 210,
		"25 WPM",  48, 144,  48,  96, 192,
		"27 WPM",  45, 133,  45,  88, 179,
		"30 WPM",  40, 120,  40,  80, 160,
		"35 WPM",  34, 103,  34,  69, 138,
		"40 WPM",  30,  90,  30,  60, 120,
		"50 WPM",  24,  72,  24,  48,  96,
		"75 WPM",  16,  48,  16,  32,  64};

/*		"24 WPM",  50, 150,  50, 100, 200,  */
/*		"26 WPM",  47, 138,  47,  91, 185,  */
/*		"28 WPM",  43, 129,  43,  86, 172,  */
/*		"29 WPM",  41, 124,  41,  83, 166,  */
/*		"30 WPM",  40, 120,  40,  80, 160,  */


/* The m-code is for LHL7 coding/decoding. Each long starts with a 1 if the
 * character is available, and procedes 0=dot, 1=dash, 2=x2, 3=x3, 4=x4, 5=x5.
 * The corresponding ASCII character comes from the array index+32. Quite
 * a few aren't used and are 0's.
 */
long int   m[64]=
/*      sp               "               $                       '    */
    {   1130,   0000,   0000,   0000,   0000,   0000,   0000,   0000,
 
/*       (       )                       '       -       .       /    */
        0000,   0000,   0000,   0000,   0000,    1,  1010101,1120212, 

/*       0       1       2       3       4       5       6       7    */
         115,   1014,  10213,  10312,   1041,    105,   1104,  11203, 

/*       8       9       :       ;               =               ?    */
       11302,   1140,   0000,   0000,   0000,   0000,   0000,   0000, 

/*               A       B       C       D       E       F       G    */
        0000,    101,   1103,  11010,   1102,     10,  10210,   1120, 

/*       H       I       J       K       L       M       N       O    */
         104,    102,   1013,   1101,  10102,    112,    110,    113, 

/*       P       Q       R       S       T       U       V       W    */
       10120,  11201,   1010,    103,     11,   1021,   1031,    1012, 

/*       X       Y       Z                               _            */
       11021,  11012,  11202,   0000,   0000,   0000,   0000,   0000};

/*************************************************************************/

/* morse_ch(c, wpm)  -  Transmits a Morse Code character at wpm speed.
 * Does not return until character is complete.
 */
void morse_ch(unsigned char c, int wpm)
	{
	int ic;

	/* The following mc[] table defines the morse code sequences
	 * in terms of bits. Starting from lsb and continuing until only one 
	 * "1" remains, send a dot for a "0" and a dash for a "1", with
	 * a right shift after each dot or dash.  */

	/*                    sp         "         $              '    */
	static int    mc[] = {0x17,0x01,0x52,0x01,0xC8,0x01,0x01,0x5E,
 
	/*                     (    )              ,    -    .    /    */
	                     0x2D,0x6D,0x01,0x01,0x73,0x61,0x6A,0x29,
	
	/*                     0    1    2    3    4    5    6    7    */
	                     0x3f,0x3e,0x3c,0x38,0x30,0x20,0x21,0x23,

	/*                     8    9    :    ;         =   KN    ?    */
	                     0x27,0x2f,0x47,0x55,0x01,0x31,0x2D,0x4C,

	/*                          A    B    C    D    E    F    G    */
	                     0x01,0x06,0x11,0x15,0x09,0x02,0x14,0x0b,

	/*                     H    I    J    K    L    M    N    O    */
	                     0x10,0x04,0x1e,0x0d,0x12,0x07,0x05,0x0f,

	/*                     P    Q    R    S    T    U    V    W    */
	                     0x16,0x1b,0x0a,0x08,0x03,0x0c,0x18,0x0e,

	/*                     X    Y    Z   AR   AS   SK   Err   _    */
        	             0x19,0x1d,0x13,0x2A,0x22,0x68,0x80,0x6C};

	if (c=='{')			/* Special case: Key down 1 second */
	    {send(1);
	     delay(999);
	     send(0);
	     return;
	    }

	if (c=='}')			/* Special case: Key up 1 second */
	    {delay(999);
	     return;
	    }

	if (c>95)  c -= 64;		/* Convert lc to caps */
	else c -= 32;			/* Move subscript to (0, 63) */
	ic = mc[(int)c];		/* Ch to morse code lookup */

	if (ic==0x17)			/* A space character */
	    {delay(t[wpm].sp);
	     return;
	    }
	while (ic!=0x01)
	    {if (ic&0x01)		/* Send a dash */
		{send(1);
		 delay(t[wpm].dash);
		 send(0);
		 delay(t[wpm].dd);	/* Dot/dash delay */
		}
	    else
		{send(1);		/* Send a dot */
		 delay(t[wpm].dot);
		 send(0);
		 delay(t[wpm].dd);	/* Dot/dash delay */
		}
	    ic >>= 1;
	    }
	delay(t[wpm].ch);		/* Delay between characters */
	}

/* morse(cmd, ichar) - CW, LHL7, PUA43, LTI transmission routines. Keeps track
 * of the send buffer and display it.  Alphanumeric keyboard entries
 * are sent here during both transmission and reception. Commands:
 * INIT_CW	0	Initialize Morse code routine
 * ADD_CH	1	Put char into cw buffer
 * SEND_CW	2	Send a character if available
 *
 * Two special characters are:
 *	# indicates the beginning and ending of a repeat string
 *	@ indicates to stop sending a repeat string
 * These two characters do not actually send any Morse code, themselves.
 * Only a @ will be accepted during a repeat.
 * The repeat string cancels all current sending activity.
 */
void morse(int cmd, int ichar)
	{
	#define BUFSIZE 179
	static int cwbuff[BUFSIZE],	/* Circular buffer for cw xmit */
		    *addptr,		/* Next location to put ch */
		    *sendptr;		/* Next location to get ch from */
	static int bufcount;		/* Ch left to be sent */
	static int cw_key_stat;		/* CW status based on right-alt key */
	int i, right_alt, nt;
	static int repeat;		/* 0 --> No repeat activity
					   1 --> Entering repeat
					   2 --> Repeat entered and active */
	static int *repeat_endptr;	/* Buff always starts at cwbuff */
	unsigned char c, *idptr;	/* For sending CW ID with FSK */
	int            remainder;
	long int       t_new;
	struct tm      *tp;
	int	       tsec;		/* tp points to public version of
					   struct tm and we need part kept
					   for use here */

	switch (cmd){
	    case INIT_CW:{
		for (i=0; i<BUFSIZE; i++)
		    cwbuff[i] = 0;
		addptr = cwbuff;
		sendptr = cwbuff;
		bufcount = 0;
		cw_key_stat = 0;
		repeat = 0;
		break;
		}
	    case ADD_CH:
		{
		if (repeat==2)
		    {
		    if (ichar==(int)'@')	/* Only key responded to */
			{
			repeat=0;
			sendptr = cwbuff;	/* Fresh start */
			addptr = cwbuff;
			bufcount = 0;
			break;
			}
		    else
			break;		/* Break case */
		    }
		if (repeat==0 && ichar==(int)'#')
		    {
		    repeat = 1;		/* Start of repeat enter */
		    sendptr = cwbuff;
		    addptr = cwbuff;
		    if (mode==CW || mode==LTI)
			print_medium('#', 255); /* Print # on screen */
		    else if (mode==LHL7)
			print_medium_1('#', 255); /* Print # on screen */
		    break;
		    }
		if (repeat==1 && ichar==(int)'#')
		    {
		    repeat = 2;		/* End of repeat enter */
		    repeat_endptr = addptr-1;
		    if (mode==CW || mode==LTI)
			print_medium('#', 255); /* Print # on screen */
		    else if (mode==LHL7)
			print_medium_1('#', 255); /* Print # on screen */
		    break;
		    }
		if (ichar==8)	/* Backspace */
		    {
		    if (bufcount <= 0)  break;	/* Don't backspace empty buf */
		    bufcount--;
		    if (--addptr < cwbuff)
			addptr = cwbuff+(BUFSIZE-1);	/* Circular buffer */
		    if (mode==CW || mode==LTI)
			print_medium(8, 254); /* Backspace display */
		    else if (mode==LHL7)
			print_medium_1(8, 254);
		    break;
		    }
		if (bufcount < BUFSIZE)	/* If there is room */
		    {
		    bufcount++;
		    *addptr = ichar;		/* Put char into buffer */
		    /* Last array position is cwbuff+BUFSIZE-1:  */
		    if (++addptr >= cwbuff+BUFSIZE)
			addptr = cwbuff;	/* Circular buffer */

		    if (mode==CW || (mode==LTI && transmit))
			print_medium((char)ichar, 255); /* Print on screen */
		    else if (mode==LHL7)
			print_medium_1((char)ichar, 255); /* Print on screen */
		    }
		else
		    beep(50);			/* No room */
		break;
		}
	    case SEND_CW:			/* CW, LTI, LHL7 or PUA43 */
	      {if (mode==CW || mode==LTI)
		{
		right_alt=_bios_keybrd(_NKEYBRD_SHIFTSTATUS) & 0x0800;
		if ( (right_alt > 0) && (cw_key_stat==0) )
		    {
		    ASYNCOUT(ESC_CHAR);  ASYNCOUT('T');
		    ASYNCOUT(0X0013); ASYNCOUT(2); /* Xmit, t stat, key dwn */ 
		    ASYNCOUT(fgainl[xmit_power]); ASYNCOUT(fgainh[xmit_power]);
		    cw_key_stat=1;
		    break;
		    }
		if ( (right_alt == 0) && (cw_key_stat==1) )
		    {
		    ASYNCOUT(ESC_CHAR);  ASYNCOUT('T');
		    ASYNCOUT(0X0003); ASYNCOUT(2);
		    ASYNCOUT(fgainl[xmit_power]); ASYNCOUT(fgainh[xmit_power]);
		    cw_key_stat=0;
		    break;
		    }
		if (repeat==1)  break;
		/* Randomization of the frequency occurs in LTI.
		 * The randomization is controlled the same as for PUA43
		 * in using a random file, but the frequency span is
		 * adjustable.
		 * LTI reception removes the frequency shift.
		 */
		if (mode==LTI && (repeat==2 || bufcount>0))
		    {
		    if (ms_available)
			t_new=update_clock(1)+dsp_clock_base;
		    else
			(void)time(&t_new);
		    if (time_type)
			{
			tp = localtime(&t_new); /* time_type==1 for local */
		        do_t_error(23, "E23 LTI w/o UT", 20);
			}
		    else
			tp = gmtime(&t_new);    /*          ==0 for gmt */

		    code_min=tp->tm_min;	/* Randomize (like pua43) */
		    code_min += (60*tp->tm_hour);
		    fsk=(int)(lti_freq()+0.5);  /* Set code freq */
		    set_fsk(1);
		    }
		if (repeat==2)
		    {
		    morse_ch((unsigned char)*sendptr, cw_speed); /* CW  char */
		    if (++sendptr > repeat_endptr)
			sendptr = cwbuff;	/* Keep repeating */
		    break;
		    }
		if (bufcount > 0)	/* If there are char to be sent */
		    {
		    bufcount--;
		    morse_ch((unsigned char)*sendptr, cw_speed); /* CW  char */
		    if (++sendptr >= cwbuff+BUFSIZE)
			sendptr = cwbuff;	/* Circular buffer */
		    }
		break;
		}			/* End SEND_CW for mode==CW, LTI */
	      else if (mode==LHL7)
		{
		right_alt=_bios_keybrd(_NKEYBRD_SHIFTSTATUS) & 0x0800;
		if ( (right_alt > 0) && (cw_key_stat==0) )
		    {
		    fsk=600; set_fsk(1);
		    ASYNCOUT(ESC_CHAR);  ASYNCOUT('T');
		    ASYNCOUT(0X0011); ASYNCOUT(2); /* Xmit, no stat, key dwn*/ 
		    ASYNCOUT(fgainl[xmit_power]); ASYNCOUT(fgainh[xmit_power]);
		    cw_key_stat=1;
		    break;
		    }
		if ( (right_alt == 0) && (cw_key_stat==1) )
		    {
		    ASYNCOUT(ESC_CHAR);  ASYNCOUT('T');
		    ASYNCOUT(0X0001); ASYNCOUT(2); /* Xmit, no stat, key up*/ 
		    ASYNCOUT(fgainl[xmit_power]); ASYNCOUT(fgainh[xmit_power]);
		    cw_key_stat=0;
		    break;
		    }
/* Not necessary? */ if(ms_available)  update_clock(0);  /* Get clock up-to-date */
		if (repeat==1)  break;
		if (repeat==2)
		    {
#if 0
 This needs to be, but not until the regular thing is wrking
		    morse_ch((unsigned char)*sendptr, cw_speed); /* CW  char */
		    if (++sendptr > repeat_endptr)
			sendptr = cwbuff;	/* Keep repeating */
		    break;
#endif
		    }
		if (bufcount > 0)	/* If there are char to be sent */
		    {
		    bufcount--;
		    for(i=0; i<lhl_code((unsigned char)*sendptr, &lc0); i++)
			{
			if (lc0.cc=='&')  /* Special case of ID */
			    {fsk=cwoffset; set_fsk(1);
			     idptr=&code_id[0];
			     while ((c=*idptr++) != 0)
				morse_ch(c,20); /* Send at 25 wpm */
			     break;
			    }
			else		/* Send a character */
			    {
			    fsk=lc0.tone[i];
			    set_fsk(1);

			    /* Wait here until the  start of symbol period */ 
			    while(1)
				{
				t_new=update_clock(1) + dsp_clock_base;
				tp=localtime(&t_new);	/* Need sec of min */

				/* Break every 2*code_nx2 sec */
				remainder=tp->tm_sec % (2*code_nx2);
				if (remainder==0)
				    break;
				}

			    /* Display the corrected time:   */
			    setviewport(90, 470, 127, 479, DO_CLIP);
			    clearviewport();
			    settextstyle(0, 0, 1);
			    settextjustify(RIGHT_TEXT, TOP_TEXT);
			    moveto(36, 0);
			    sprintf(cc, "T:%d", tp->tm_sec);
			    outtext(cc);

			    lhl7_update(tp->tm_sec, 3); /* Display tx data */
			    for (nt=0; nt<code_nx2; nt++) /* Rpt nx2 times */
			          {while (1)
			             {
				     t_new=update_clock(1) + dsp_clock_base;
				     tp=localtime(&t_new); /* Need sec of min*/
				     /* Break every 2 sec */
				     remainder=tp->tm_sec % 2;
				     if (remainder==0)
					break;
			             }
			           fsk2sec();  /* Send a tone for 2 sec */
			          }
			    }		/* End else */

			}
		    if (++sendptr >= cwbuff+BUFSIZE)  /* Move ptr to next ch */
			sendptr = cwbuff;	/* Circular buffer */
		    }			/* End of send LHL ch */
		break;
		}			/* End mode=LHL */

	      else if (mode==PUA43)
		{
		/* Wait here until the  start of symbol period */ 
		while(1)
		    {
		    if (ms_available)
			t_new=update_clock(1)+dsp_clock_base;
		    else
			(void)time(&t_new);
		    if (time_type)
			{
			tp = localtime(&t_new); /* time_type==1 for local */
		        do_t_error(24, "E24 PUA43 w/o UT", 20);
			}
		    else
			tp = gmtime(&t_new);    /*          ==0 for gmt */

		    tsec=tp->tm_sec;		/* Private copy */
		    code_min=tp->tm_min;	/* Randomize pua43 */
		    code_min += (60*tp->tm_hour);
		    remainder=tp->tm_sec % 2;	/* Break every 2 sec */
		    if (remainder==0)
			break;
		    }
		if (tsec < 56)		/* Send a PUA43 character */
		    {
		    /* Display the corrected time:   */
		    setviewport(90, 470, 127, 479, DO_CLIP);
		    clearviewport();
		    settextstyle(0, 0, 1);
		    settextjustify(RIGHT_TEXT, TOP_TEXT);
		    moveto(36, 0);
		    sprintf(cc, "T:%d", tsec);
		    outtext(cc);

		    i=tsec/2;
		    /* Repeat first 14 if pua14_28==FALSE */
		    if (pua14_28==FALSE && tsec >= 28)
			i=(tsec-28)/2;
		    c=pua43_xstring[i];	/* Get string member */
		    fsk=(int)(pua43_freq(c)+0.5);  /* Set code freq */
		    set_fsk(1);

		    /* Display the transmit status:      */
		    pua43_update(tsec, 4);

		    fsk2sec();		/* Send a tone for 2 sec */
		    break;		/* Break the switch(cmd) */
		    }
		else if (tsec==56)
		    {
		    fsk=400; set_fsk(1);
		    idptr=&code_id[0];
		    while ((c=*idptr++) != 0)
			morse_ch(c,20); /* Send at 25 wpm */
		    break;
		    }
		else			/* Kill 1.8 sec */
		    {
		    delay(450); update_clock(0);
		    delay(450); update_clock(0);
		    delay(450); update_clock(0);
		    delay(450); update_clock(0);
		    break;
		    }
		}			/* End mode==PUA43  */
		break;
	      }				/* End SEND_CW (or code tones ) */
	    default: break;
	    }				/* End switch(cmd) */
	}



/*  LHL7_UPDATE()  -  Accumulates data for FSK coding per W7LHL
 *  Data is received in groups of 3 that need to be integrated together.
 *  These ae identified by the time_tag, which is the number of the
 *  2 second interval, (0, 29) multiplied by 8 and a sequence number,
 *  0, 1 or 2 added on.
 *  The integer code_nx2 can be 1, 2, 5, 10, or 30 and specifies the number
 *  of 2 second intervals to be integrated together.
 *  The integer lhl_bins can be 1, 3, 5, 7, or 9 and is the number of
 *  spectral bins to be examined for each tone frequency.  In addition,
 *  on either side of the bins being examined there is a guard bin and then
 *  two noise measurement bins.
 *  peak_ave is 0 if the largest power  of lhl_bins is taken,
 *  and 1 if the average power over all lhl_bins is used.
 *
 *  what:   -1  Reset and zero sums.
 *           0  Close status box, but leave sums as they are.
 *           1  Open status box, use sums as they are.
 *           2  Update all displays.
 *	     3  Update info-box during transmit
 */
void lhl7_update(int time_tag, int what)
	{
	char cc[600];
	double big;
	int i, ii, bigi;
	int tx2;			/* Time tag multiple of 2 sec */
	int still_possible;		/* To see if mcbyte going anywhere */
	double lhl_n[7];
	double lhl_spn[7];
	double lhl_snr[7];
	static int mc_cnt=0;
	static double lhl_pwr[111];
	static long int   mcbyte=1;
	long int          mtest;

	if (what==-1)			/* Initialize LHL7 decode */
	   {
	   for (i=0; i<111; i++)  lhl_pwr[i]=0.0;
	   }

	else if (what==0)		/* Close LHL7 status box, no reset */
	   {
	   lb_close(&lb_lhl7, 0);	/* Undisplay the box */
	   current_info_box=0;		/* Tell everybody */
	   }
	else if (what==1)		/* Open LHL7 data box */
	   {
	    if (current_info_box  &&  (current_info_box != &lb_lhl7))
		lb_close(current_info_box, 0);
	    lb_create(&lb_lhl7, "LHL 7-TONE",
		 482, 148, 638, 340, 0, NON_MODAL);
	   }
	else if (what==2)		/* Update power sums and update disp */
	   {
	    /* LHL7 extends from 600 Hz to 1500 Hz. In order to have a up to 9
	     * signal bins, a guard bin on either side and 4 noise bins we need
	     * 7 bins below 600 Hz (down to 57) and 7 bins above 1500 Hz (up
	     * to bin 167). For simplicity this collection of 111 bins will
	     * all be accumulated, even though some will not be used.
	     */
	    for (i=57; i<=167; i++)
		{
		if (bits_8_16)
		  {
		  lhl_pwr[i-57] += pd16[i];
		  }
		else
		  {
	    	  lhl_pwr[i-57] += power10[raw_data[i]];
		  }
		}

	    /* Now, examine signal and noise bins to determine the s+n/n for
	     * each of the 7 signalling frequencies. Two options are involved:
	     *    1 - The number of bins per frequency to be examined,
	     *           lhl_bin=1, 3, 5, 7 or 9.
	     *    2 - The criteria to be applied over the bins, either maximum
	     *           observed value (lhl7_peak_ave=0), or the ave of the
	     *           bin powers (lhl7_peak_ave=1).
	     *
	     * To find when to process:
	     * First, bits 0, 1 are the sequence number, 0, 1 or 2,
	     * and we need the last, or 2.  Then bits 2 to 7 of time_tag
	     * have the time (bit 2 value is 1 sec), in steps of 2 seconds,
	     * since the top of the minute, i.e., bit 3
	     * alternates 0/1 every 2 sec.  If we AND off
	     * these bits, shift the 2 second bit into lsb, and then
	     * integer divide by code_nx2, the remainder is 0 at the first
	     * two second group, and progresses to code_nx2-1 at the last
	     * group, which is what we need to process the data.
	     */
	    if ( (time_tag & 0X0003)==2 &&
	        (((time_tag & 0X00F8)>>3) % code_nx2)==(code_nx2-1) )
	        {
	        for (i=0; i<7; i++)
	          {switch(lhl_bins)
		    {
		    case 1:
		        for (i=0; i<7; i++)
			    {
			    /* Look above about bin#. 7 --> 600Hz */
			    /* Note: peak & ave same if lhl_bins==1 */
			    lhl_n[i]=0.25*(lhl_pwr[4+16*i]+lhl_pwr[5+16*i]+
				lhl_pwr[9+16*i]+lhl_pwr[10+16*i]);
			    lhl_spn[i]=lhl_pwr[7+16*i];
			    lhl_snr[i]=lhl_spn[i] / lhl_n[i];
			    }
		        break;
		    case 3:
			for (i=0; i<7; i++)
			    {
			    lhl_n[i]=0.25*(lhl_pwr[3+16*i]+lhl_pwr[4+16*i]+
				lhl_pwr[10+16*i]+lhl_pwr[11+16*i]);
			    if (lhl_peak_ave)  /* Using average */
				{
				lhl_spn[i]=0.3333333333*(lhl_pwr[6+16*i]+
				    lhl_pwr[7+16*i]+lhl_pwr[8+16*i]);
				}
			    else	/* If looking for highest */
				{
				lhl_spn[i]=lhl_pwr[6+16*i];
				if (lhl_pwr[7+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[7+16*i];
				if (lhl_pwr[8+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[8+16*i];
				}
			    lhl_snr[i]=lhl_spn[i] / lhl_n[i];
			    }
			break;
		    case 5:
			for (i=0; i<7; i++)
			    {
			    lhl_n[i]=0.25*(lhl_pwr[2+16*i]+lhl_pwr[3+16*i]+
				lhl_pwr[11+16*i]+lhl_pwr[12+16*i]);
			    if (lhl_peak_ave)  /* Using average */
				{
				lhl_spn[i]=0.2*(lhl_pwr[5+16*i]+
				    lhl_pwr[6+16*i]+lhl_pwr[7+16*i]+
				    lhl_pwr[8+16*i]+lhl_pwr[9+16*i]);
				}
			    else	/* If looking for highest */
				{
				lhl_spn[i]=lhl_pwr[5+16*i];
				if (lhl_pwr[6+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[6+16*i];
				if (lhl_pwr[7+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[7+16*i];
				if (lhl_pwr[8+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[8+16*i];
				if (lhl_pwr[9+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[9+16*i];
				}
			    lhl_snr[i]=lhl_spn[i] / lhl_n[i];
			    }
			break;
		    case 7:
			for (i=0; i<7; i++)
			    {
			    lhl_n[i]=0.25*(lhl_pwr[1+16*i]+lhl_pwr[2+16*i]+
				lhl_pwr[12+16*i]+lhl_pwr[13+16*i]);
			    if (lhl_peak_ave)  /* Using average */
				{
				lhl_spn[i]=0.142857143*(lhl_pwr[4+16*i]+
				    lhl_pwr[5+16*i]+lhl_pwr[6+16*i]+
				    lhl_pwr[7+16*i]+lhl_pwr[8+16*i]+
				    lhl_pwr[9+16*i]+lhl_pwr[10+16*i]);
				}
			    else	/* If looking for highest */
				{
				lhl_spn[i]=lhl_pwr[4+16*i];
				if (lhl_pwr[5+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[5+16*i];
				if (lhl_pwr[6+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[6+16*i];
				if (lhl_pwr[7+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[7+16*i];
				if (lhl_pwr[8+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[8+16*i];
				if (lhl_pwr[9+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[9+16*i];
				if (lhl_pwr[10+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[10+16*i];
				}
			    lhl_snr[i]=lhl_spn[i] / lhl_n[i];
			    }
			break;
		    case 9:
			for (i=0; i<7; i++)
			    {
			    lhl_n[i]=0.25*(lhl_pwr[16*i]+lhl_pwr[1+16*i]+
				lhl_pwr[13+16*i]+lhl_pwr[14+16*i]);
			    if (lhl_peak_ave)  /* Using average */
				{
				lhl_spn[i]=0.111111111*(lhl_pwr[3+16*i]+
				    lhl_pwr[4+16*i]+lhl_pwr[5+16*i]+
				    lhl_pwr[6+16*i]+lhl_pwr[7+16*i]+
				    lhl_pwr[8+16*i]+lhl_pwr[9+16*i]+
				    lhl_pwr[10+16*i]+lhl_pwr[11+16*i]);
				}
			    else	/* If looking for highest */
				{
				lhl_spn[i]=lhl_pwr[3+16*i];
				if (lhl_pwr[4+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[4+16*i];
				if (lhl_pwr[5+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[5+16*i];
				if (lhl_pwr[6+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[6+16*i];
				if (lhl_pwr[7+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[7+16*i];
				if (lhl_pwr[8+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[8+16*i];
				if (lhl_pwr[9+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[9+16*i];
				if (lhl_pwr[10+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[10+16*i];
				if (lhl_pwr[11+16*i] > lhl_spn[i])
				    lhl_spn[i]=lhl_pwr[11+16*i];
				}
			    lhl_snr[i]=lhl_spn[i] / lhl_n[i];
			    }
			break;
		    default:
		        do_t_error(20, "E20 LHL7 # Bins", 20);
		        break;
		    }			/* End, switch(lhl_bins) */
		  }			/* End, find highest snr */

	        /* Zero the sums for the next symbol */
	        for (i=0; i<=110; i++)
		  {
		  lhl_pwr[i] = 0.0;
		  }

	        /* To decode the signal, we need to find the highest snr of
	           the 7. */
	        big=lhl_snr[0];
	        bigi=0;
	        for(i=1; i<7; i++)
		  {if (lhl_snr[i] > big)
		     {
		     big=lhl_snr[i];
		     bigi=i;
		     }
		  }
	        /* Now try to construct mcbyte: */

		/* I believe this is redundant with the following check */
	        if (mc_cnt>6)		/* They don't come that size so reset*/
		   {
		   mc_cnt=0;
		   mcbyte=1;		/* This becomes leading 1 */
		   }

		/* Check if it is ever going to be possible to create a char */
		still_possible=FALSE;
		for (i=0; i<64; i++)
		    {
		    mtest=m[i];
		    while (2*mcbyte < mtest)  mtest = mtest/10;
		    if (mtest==mcbyte)
			{
			still_possible=TRUE;
			break;
			}
		    }
		if (!still_possible)
		    {
		    mc_cnt=0;
		    mcbyte=1;
		    }

		/* To start a character, there is a 1 in the MS position to
		 * delimit the start. Then the next symbol must be a 0, 1,
		 * or 2.
		 */
		if (mcbyte>1 || bigi==0 || bigi==1 || bigi==2)
		    {
	            mc_cnt++;
	            switch(bigi)
		       {case 0: /* An end of character */
			   /* Convert to character and print */
			   for (i=0; i<64; i++)   /* Search the whole table */
			      {if (mcbyte==m[i])  /* There is a match */
				  {
				  if (lhl7_2line) /* 2 display configurations*/
				      print_medium_2(i+32, 255);
				  else            /* one line display */
				      print_big(i+32, 255, c_text_bold);
				  break;
				  }
			      }
			   mc_cnt=0;
			   mcbyte=1;
			   break;
		        case 1: mcbyte=10*mcbyte;	/* A dot (=0) */
			     break;
		        case 2: mcbyte=10*mcbyte + 1;	/* A dash (=1) */
			     break;
		        case 3: mcbyte=10*mcbyte + 2;	/* Mult x2 */
			     break;
	 	        case 4: mcbyte=10*mcbyte + 3;	/* Mult x3 */
			     break;
		        case 5: mcbyte=10*mcbyte + 4;	/* Mult x4 */
			     break;
		        case 6: mcbyte=10*mcbyte + 5;	/* Mult x5 */
			     break;
		       }			/* End switch(bigi) */
		    }				/* End we have dot or dash */

		if (current_info_box == 0)  return;
	        lb_open(&lb_lhl7, 0);		/* Draw the box */

	        /* Fixed text in LHL7 box:     */
	        lb_line(&lb_lhl7, 2,  TXT, "Symbol Hz  S+N/N dB",  0);
	        lb_line(&lb_lhl7, 15, TXT, "Close Box=Alt A ", 0);
	        lb_line(&lb_lhl7, 16, TXT, "Change Params=Alt B", 0);
	        lb_line(&lb_lhl7, 17, TXT, ".                 .", 0);

		sprintf(cc+25, " %d Bins per Tone", lhl_bins);
		lb_line(&lb_lhl7, 0, BOLD, cc+25, 0);
		if (lhl_peak_ave)
		    lb_line(&lb_lhl7, 1, BOLD, " Average Over Bins", 0);
		else
		    lb_line(&lb_lhl7, 1, BOLD, " Peak of Bins", 0);

		sprintf(cc+75,  " EOC  600  %7.3lf", p2db(lhl_snr[0]));
		lb_line(&lb_lhl7, 3, BOLD, cc+75, 0);
		sprintf(cc+100, " Dot  750  %7.3lf", p2db(lhl_snr[1]));
		lb_line(&lb_lhl7, 4, BOLD, cc+100, 0);
		sprintf(cc+125, " Dash 900  %7.3lf", p2db(lhl_snr[2]));
		lb_line(&lb_lhl7, 5, BOLD, cc+125, 0);
		sprintf(cc+150, " x2  1050  %7.3lf", p2db(lhl_snr[3]));
		lb_line(&lb_lhl7, 6, BOLD, cc+150, 0);
		sprintf(cc+175, " x3  1200  %7.3lf", p2db(lhl_snr[4]));
		lb_line(&lb_lhl7, 7, BOLD, cc+175, 0);
		sprintf(cc+200, " x4  1350  %7.3lf", p2db(lhl_snr[5]));
		lb_line(&lb_lhl7, 8, BOLD, cc+200, 0);

		sprintf(cc+225, " x5  1500  %7.3lf", p2db(lhl_snr[6]));
		lb_line(&lb_lhl7, 9, BOLD, cc+225, 0);

		sprintf(cc+275, " Tag: %d", (0X00FC & time_tag)>>2);
		lb_line(&lb_lhl7, 11, BOLD, cc+275, 0);

		sprintf(cc+300, " bigi = %d", bigi);
		lb_line(&lb_lhl7, 12, BOLD, cc+300, 0);

		sprintf(cc+325, " mcbyte = %ld", mcbyte);
		lb_line(&lb_lhl7, 13, BOLD, cc+325, 0);

		if (current_info_box==&lb_lhl7)
	           lb_update(&lb_lhl7);	/* Redraw the info box with new data */
	        }			/* End, if (time for update) */
	    }				/* End, if (what==2) */
	else if (what==3)		/* Update display, during transmit */
	   {
	   if (current_info_box == 0)  return;
	        lb_open(&lb_lhl7, 0);		/* Draw the box */

	   /* Fixed text in LHL7 box:     */
	   lb_line(&lb_lhl7, 2,  TXT, "Symbol Tone",  0);
	   lb_line(&lb_lhl7, 15, TXT, "Close Box=Alt A ", 0);
	   lb_line(&lb_lhl7, 16, TXT, "Change Params=Alt B", 0);
	   lb_line(&lb_lhl7, 17, TXT, ".                 .", 0);

	   lb_line(&lb_lhl7, 0, BOLD, " ", 0);
	   lb_line(&lb_lhl7, 1, BOLD, " ", 0);

	   lb_line(&lb_lhl7, 3, BOLD, " EOC  600 Hz", 0);
	   lb_line(&lb_lhl7, 4, BOLD, " Dot  750 Hz", 0);
	   lb_line(&lb_lhl7, 5, BOLD, " Dash 900 Hz", 0);
	   lb_line(&lb_lhl7, 6, BOLD, " x2  1050 Hz", 0);
	   lb_line(&lb_lhl7, 7, BOLD, " x3  1200 Hz", 0);
	   lb_line(&lb_lhl7, 8, BOLD, " x4  1350 Hz", 0);
	   lb_line(&lb_lhl7, 9, BOLD, " x5  1500 Hz", 0);

	   sprintf(cc+275, " Current Tone=%d", fsk);
	   lb_line(&lb_lhl7, 11, BOLD, cc+275, 0);

	   sprintf(cc+300, " Current Char=%c", lc0.cc);
	   lb_line(&lb_lhl7, 12, BOLD, cc+300, 0);
	   sprintf(cc+325, " Symbol Time=%d", time_tag);
	   lb_line(&lb_lhl7, 13, BOLD, cc+325, 0);

	   if (current_info_box==&lb_lhl7)
	          lb_update(&lb_lhl7);	/* Redraw the info box with new data */

	   }		/* End if what==3 */
	}

/*  PUA43_UPDATE()  -  Accumulates data for FSK coding per W7PUA.
 *  Data is received in groups of 3 that need to be integrated together.
 *  This is two fixed length messages, of 14 or 28 positions, that are
 *  repeated over and over, either in 30 or 60 seconds.
 *  Integration improves the message quality!
 *  If pua14_28==0 there are 14 positions, if ==1, there are 28.
 *
 *  what:   -3  Allocates arrays in memory
 *          -2  De allocates these arrays
 *          -1  Reset and zero sums.
 *           0  Close status box, but leave sums as they are.
 *           1  Open status box, use sums as they are.
 *           2  Update s/n estimates and all displays.
 *           3  Update noise estimates only - use when idling
 *           4  Update status box for transmit
 *  Note: what==-3 and what==-2 are to be called only when graphics
 *  are not being used.
 *
 *  Redone for real-time noise measurement, 2 May 00  RSL
 */
void pua43_update(int time_tag, int what)    /* nseq)  */
	{
	char bigi, bigi2, bigi3, bigi4, bigi5;
	char cc[600];
	int savbigi, savbigi2, savbigi3, savbigi4;
	double big, big2, big3, big4, big5;
	static int pua43_count=0;
	int i, i43, j, qint;
	int ipile;			/* Which of 14/28 positions */
	int nseq;
	int tt;
int rawtt;
	static FILE* rn_file;

	/* The array:  double pua43_pwr[28][43]
	 * is the accumulated power for each character
	 * position at each of the 43 frequencies.  It is reset only by
	 * a mode change that does a call to this function with what=-1.
	 */
	static double* pua43_pwr=0;

	/* Noise is assumed to be independent of the character position, but
	 * not freq, and so is integrated in a single dimensional array:
	 * double pua43_n[43];
	 */
	static double* pua43_n=0;

	/* Noise measurements are RC decayed by the global variable
	 * pua43_noisedecay (just under 1). The new measurement is added in
	 * as noisefraction.
	 */
	double noisefraction;

	/* The noise measurement is used to generate a normalization factor
	 * that is used to scale the signal measurement. If the noise is
	 * measured as being high (pnorm[i]) then the signal at that freq
	 * will be scaled back. This is inverse of the ratio of
	 * pua43_n[i] to the average over all 43.
	 */
	static double* pnorm=0;

	double pua43_n_sum, pua43_n_sumsq;	/* Get statistical param */
	double pua43_n_sigma;
	double pua43_s_sum, pua43_s_sumsq;
	double pua43_s_sigma;
	double x;
	char savch;
	static int mem_available=FALSE;  /* TRUE if arrays exist */

	char cnvt43(char);

	if (what==-3)			/* Initialize & allocate data space */
	    {
	    if ( (pua43_pwr=(double *)calloc(1204, sizeof(double)) ) == NULL)
		{printf("Unable to get memory for pua43_pwr[].\n");
		 printf("Use mode_mask in uhfa.cfg to disable PUA43\n");
		 printf("'Enter' to continue\n");
		 getch();
		 exit(0);
		}
	    if ( (pua43_n = (double *)calloc(43, sizeof(double)) ) == NULL)
		{printf("Unable to get memory for pua43_n[].\n");
		 printf("Use mode_mask in uhfa.cfg to disable PUA43\n");
		 printf("'Enter' to continue\n");
		 getch();
		 exit(0);
		}
	    if ( (pnorm = (double *)calloc(43, sizeof(double)) ) == NULL)
		{printf("Unable to get memory for pnorm[].\n");
		 printf("Use mode_mask in uhfa.cfg to disable PUA43\n");
		 printf("'Enter' to continue\n");
		 getch();
		 exit(0);
		}
	    /* The array pua43_rnd[] is a set of 1440 random integers (0,42)
	     * that stir the frequency selection, based on the minute of the
	     * day. These are global, as they are needed for both transmit
	     * and receive.
	     */
	    if ( (pua43_rnd = (int *)calloc(1440, sizeof(int)) ) == NULL)
		{printf("Unable to get memory for pua43_rnd[].\n");
		 printf("Use mode_mask in uhfa.cfg to disable PUA43\n");
		 printf("'Enter' to continue\n");
		 getch();
		 exit(0);
		}
	    mem_available=TRUE;

	    if (NULL == ( rn_file=fopen("UHFA_43A.RND", "r") ))
		{
		printf("\nNo UHFA_43A.RND random number file was found.");
		printf(" This must be in the the same\ndirectory as UHFA.EXE");
		printf(" file. Correct by adding file, or disabling the");
		printf(" PUA43\nmode by mode_mask in the UHFA.CFG file.");
		printf(" Exiting program.");
		exit(1);
		}
	    i=0;
	    while (fgets(cc, 18, rn_file) && i<1440)
		{
		sscanf(cc, " %d", &pua43_rnd[i]);
		if (pua43_rnd[i]<0 || pua43_rnd[i]>42)
		    {
		    printf("\nIllegal entry in PUA43 random number file.");
		    printf("\nCorrect the entries (0, 42), or disable the");
		    printf(" PUA43\nmode by mode_mask in the UHFA.CFG file.");
		    printf(" Exiting program.");
		    exit(1);
		    }
		i++;
		}
	    if (i != 1440)
		{
		printf("\nUHFA_43A.RND has only %d random numbers;");
		printf(" all being set to 0.\n", i);
		printf("Hit Enter to continue\n");
		getch();
		for (i=0; i<1440; i++) pua43_rnd[i]=0;
		}
	    fclose(rn_file);

	    }
	else if (what==-2)		/* Release data space */
	    {
	    if (mem_available)
		{
		free(pua43_pwr);    /* [28][43]; */
		free(pua43_n);
		free(pnorm);
		free(pua43_rnd);
		mem_available=FALSE;
		}
	    }
	else if (what==-1)		/* Reset by zeroing powers */
	    {
	    if(mem_available)
		{
		for (i=0; i<43; i++)
		    {for (j=0; j<28; j++)
			pua43_pwr[43*j+i]=0.0; /* Signal+Noise estimates */
		    pua43_n[i]=0.0;	/* Noise Estimates */
		    }
		}
	    for (j=0; j<28; j++)	/* Clear the display */
		{
		print_big(32, j, c_text);
		print_super_row(32, j, c_text);
		}
	    pua43_count=0;
	    }
	else if (what==0)		/* Close PUA43 status box, no reset */
	    {
	    lb_close(&lb_pua43, 0);	/* Undisplay the box */
	    current_info_box=0;		/* Tell everybody */
	    }
	else if (what==1)		/* Open PUA43 data box */
	    {
	    if (current_info_box  &&  (current_info_box != &lb_pua43))
		lb_close(current_info_box, 0);
	    lb_create(&lb_pua43, "PUA 43-TONE",
		 482, 148, 638, 340, 0, NON_MODAL);
	    current_info_box=&lb_pua43;
	    /* Fixed text in PUA43 box:     */
	    lb_line(&lb_pua43, 16, TXT, "Close=Alt A  Clr=^W", 0);
	    lb_line(&lb_pua43, 17, TXT, "Change Params=Alt B", 0);
	    lb_open(&lb_pua43, 0);
	    }
	else if (what==2)		/* Update power sums and update disp */
	    {
	    if (!mem_available)
		{do_t_error(21, "E21 PUA43 Memory", 200);
		 return;
		}
	    /* Use time_tag to determine which of 28/14 piles to add to.
	     * If 28 pile pua43_pwr[tt][freq]  If 14, pile pua43_pwr[tt][freq]
	     * or [tt-14][freq] if tt>=14.
	     */
		rawtt=time_tag;
/* 	    nseq=0X0003 & time_tag;	<< Not used now */
	    tt=(0X00FC & time_tag) >> 3;
	    if (pua14_28)		/* there are 28 positions */
		ipile=tt;
	    else if (!pua14_28 && tt<14) /* First 30 sec of 14 pos */
		ipile=tt;
	    else			/* second 30 sec of 14 pos*/
		ipile=tt-14;
 if (ipile<0) do_t_error(22, "E22 PUA43 t Tag", 20);
	 if (tt<28 && ipile>=0)		/* Ignore last 4 sec of minute */
	    {
	    /* PUA43 extends from 450 Hz to 2025 Hz (fft_bw==2)
	     *                    450 Hz to 1237.5 Hz (fft_bw==1)
	     *                    450 Hz to 843.75 Hz (fft_bw==0)
	     * The offset frequencies are 450 + 4*i*binsize Hz, for i= (0, 42)
	     * where binsize==9.375, 4.6875 and 2.34375 for the 3 fft_bw.
	     * For each signal bin, there is a guard bin on either side
	     * and then 1 noise bin.
	     * First update the noise estimates.
	     */
	    pua43_n_sum=0.0;
	    pua43_n_sumsq=0.0;
	    noisefraction=1.0-pua43_noisedecay;
	    for (i=0; i<43; i++)
		{
		/* RC decayed sum of noise for each of 43 freq:  */
		if (bits_8_16)
		  {
		  pua43_n[i]=
		    noisefraction*(pd16[b0[fft_bw]-2+4*i]
		    +pd16[b0[fft_bw]+2+4*i])
		    +pua43_noisedecay*pua43_n[i];
		  }
		else
		  {
	    	  pua43_n[i]=noisefraction*
		    (power10[raw_data[b0[fft_bw]-2+4*i]]
		    +power10[raw_data[b0[fft_bw]+2+4*i]])+
		    +pua43_noisedecay*pua43_n[i];
		  }
		pua43_n_sum += pua43_n[i];
		pua43_n_sumsq += (pua43_n[i]*pua43_n[i]);
		}
	    pua43_n_sum /= 43.0;
	    /* Find std dev of noise measure. This will give a measure of the
	     * quality of the message estimate.
	     */
	    x=(pua43_n_sumsq/43.0)-pua43_n_sum*pua43_n_sum;
	    if (x>0.0)
		pua43_n_sigma=sqrt(x);
	    else
		pua43_n_sigma=0.0;

	    /* Normalize to noise power sum. This estimates the noise
	     * spectrum, and on the ave, =1.0.
	     */
	    for (i=0; i<43; i++)
		{
		pnorm[i]=pua43_n_sum/pua43_n[i];
		}
	    /* Normalize the S+N and add it to the running sums. Here is where
	     * the random frequency is removed, by using i43.
	     */
	    for (i=0; i<43; i++)
		{
		if (pua43_dorandom)
		    {
		    i43=i+pua43_rnd[code_min]; /* Shift for randomizer */
		    if (i43>42)   i43 -= 43;   /* Modulus 43 */
		    }
		else
		    i43=i;			/* If random not being used */

		if (bits_8_16)
		  {
		  pua43_pwr[43*ipile+i43] += pnorm[i]*pd16[b0[fft_bw]+4*i];
		  }
		else
		  {
	    	  pua43_pwr[43*ipile+i43] +=
			pnorm[i]*power10[raw_data[b0[fft_bw]+4*i]];
		  }
		}
/* <<<<<<<<<<< */
/* ONLY DO THE FOLLOWING WHEN nseq IS 2 ?? */
	    pua43_count++;
	    /* Now find strongest of 43 for the time slot "ipile" and
	     * the statistics:   */
	    big=pua43_pwr[43*ipile+0];
	    bigi=0;
	    pua43_s_sum=0.0;
	    pua43_s_sumsq=0.0;
	    for(i=1; i<43; i++)
		{
		if (pua43_pwr[43*ipile+i]>big)
		    {big=pua43_pwr[43*ipile+i];
		     bigi=(char)i;
		    }
		pua43_s_sum += pua43_pwr[43*ipile+i];
		pua43_s_sumsq += (pua43_pwr[43*ipile+i]*pua43_pwr[43*ipile+i]);
		}
	    pua43_s_sum /= 43.0;
	    x=(pua43_s_sumsq/43.0)-pua43_s_sum*pua43_s_sum;
	    if (x>0.0)
		pua43_s_sigma=sqrt(x);
	    else
		pua43_s_sigma=0.0;
	    savbigi=(int)bigi;

	    /* Find second strongest of 43 for the time slot "ipile":   */
	    big2=0.0;
	    bigi2=0;
	    for(i=0; i<43; i++)
		{
		if (i!=savbigi && pua43_pwr[43*ipile+i]>big2)
		    {big2=pua43_pwr[43*ipile+i];
		     bigi2=(char)i;
		    }
		}
	    savbigi2=(int)bigi2;

	    /* Find third strongest of 43 for the time slot "ipile":   */
	    big3=0.0;
	    bigi3=0;
	    for(i=0; i<43; i++)
		{
		if (i!=savbigi && i!=savbigi2 && pua43_pwr[43*ipile+i]>big3)
		    {big3=pua43_pwr[43*ipile+i];
		     bigi3=(char)i;
		    }
		}
	    savbigi3=(int)bigi3;

	    /* Find fourth strongest of 43 for the time slot "ipile":   */
	    big4=0.0;
	    bigi4=0;
	    for(i=0; i<43; i++)
		{
		if (i!=savbigi && i!=savbigi2 && i!=savbigi3
			 && pua43_pwr[43*ipile+i]>big4)
		    {big4=pua43_pwr[43*ipile+i];
		     bigi4=(char)i;
		    }
		}
	    savbigi4=(int)bigi4;

	    /* Find 5th strongest of 43 for the time slot "ipile":   */
	    big5=0.0;
	    bigi5=0;
	    for(i=0; i<43; i++)
		{
		if (i!=savbigi && i!=savbigi2 && i!=savbigi3 && i!=savbigi4
			&& pua43_pwr[43*ipile+i]>big5)
		    {big5=pua43_pwr[43*ipile+i];
		     bigi5=(char)i;
		    }
		}

	    /* Change the brightness of the printing, based on confidence: */
	    if (pua43_pwr[43*ipile+savbigi] <
		    pua43_s_sum+pua43_colorscale*pua43_s_sigma)
		print_big(cnvt43(bigi), ipile, c_bar_dark);
	    else if(pua43_pwr[43*ipile+savbigi] <
		    pua43_s_sum+2*pua43_colorscale*pua43_s_sigma)
		print_big(cnvt43(bigi), ipile, c_text);
	    else
		print_big(cnvt43(bigi), ipile, c_text_bold);

	    /* Change the brightness of the printing, based on confidence: */
	    if (pua43_pwr[43*ipile+savbigi2] < 
		    pua43_s_sum+pua43_colorscale*pua43_s_sigma)
		print_super_row(cnvt43(bigi2), ipile, c_bar_dark);
	    else if(pua43_pwr[43*ipile+savbigi2] < 
		    pua43_s_sum+2.0*pua43_colorscale*pua43_s_sigma)
		print_super_row(cnvt43(bigi2), ipile, c_text);
	    else
		print_super_row(cnvt43(bigi2), ipile, c_text_bold);

	    setviewport(129, 31, 638, 33, DO_CLIP);
	    clearviewport();
	    line(16*ipile+2, 1, 16*ipile+15, 1);
	    
	    if (current_info_box == 0)  return;
	    lb_open(&lb_pua43, 0);		/* Draw the box */

	    if (pua14_28)
	        lb_line(&lb_pua43, 1, BOLD, " 28 Char total", 0);
	    else
	        lb_line(&lb_pua43, 1, BOLD, " 14 Char total", 0);

	    if (pua14_28)
		{
	        sprintf(cc+50, "Pts=%d  %.1lfdB", (14+pua43_count)/28,
		              0.5*p2db((double)((14+pua43_count)/28)));
		}
	    else
		{
	        sprintf(cc+50, "Pts=%d  %.1lfdB", (7+pua43_count)/14,
		              0.5*p2db((double)((7+pua43_count)/14)));
		}
	    lb_line(&lb_pua43, 2, BOLD, cc+50, 0);

	    /* These lines may be picked up from transmit, so clear here. */
	    lb_line(&lb_pua43, 3, BOLD, " ", 0);
	    lb_line(&lb_pua43, 4, BOLD, " ", 0);

	    sprintf(cc+125, "Pile=%d", ipile);
	    lb_line(&lb_pua43, 5, BOLD, cc+125, 0);

	    sprintf(cc+150, "Ch 1-5: %c %c %c %c %c",
		cnvt43(bigi), cnvt43(bigi2), cnvt43(bigi3),
		cnvt43(bigi4), cnvt43(bigi5));
	    lb_line(&lb_pua43, 6, BOLD, cc+150, 0);

	    /* This line may be picked up from transmit, so clear here. */
	    lb_line(&lb_pua43, 7, BOLD, " ", 0);

	    sprintf(cc+200, "Noise Sig/Av=%.4lf", pua43_n_sigma/pua43_n_sum);
	    lb_line(&lb_pua43, 8, BOLD, cc+200, 0);

	    sprintf(cc+225, "S+N/ave=%.3lf",
		pua43_pwr[43*ipile+savbigi]/pua43_s_sum);
	    lb_line(&lb_pua43, 9, BOLD, cc+225, 0);

	    sprintf(cc+250, "sigma/ave=%.4lf", pua43_s_sigma/pua43_s_sum);
	    lb_line(&lb_pua43, 10, BOLD, cc+250, 0);

	    /* This line may be picked up from transmit, so clear here. */
	    lb_line(&lb_pua43, 11, BOLD, " ", 0);

	    if (pua43_dorandom)
		sprintf(cc+300, "Random stir=%d", pua43_rnd[code_min] );
	    else
		sprintf(cc+300, "Not Randomized");
	    lb_line(&lb_pua43, 12, BOLD, cc+300, 0);

	    savch=pua43_xstring[14];
	    pua43_xstring[14]=0;
	    sprintf(cc+325, "|%s|", pua43_xstring);
	    lb_line(&lb_pua43, 13, BOLD, cc+325, 0);

	    pua43_xstring[14]=savch;
	    if (pua14_28)
		sprintf(cc+350, "|%s|", &pua43_xstring[14]);
	    else
		sprintf(cc+350, " ");
	    lb_line(&lb_pua43, 14, BOLD, cc+350, 0);

	    if (current_info_box==&lb_pua43)
	       lb_update(&lb_pua43);	/* Redraw the info box with new data */

	    }  /* if <28 */
	  }    /* if what==2 */

	else if (what==3)		/* Update noise estimates, only */
	    {
	    if (!mem_available)
		{do_t_error(21, "PUA43 Memory", 200);
		 return;
		}

	    /* For each signal bin, there is a guard bin on either side
	     * and then 1 noise bin. Update the noise estimates:
	     */
	    pua43_n_sum=0.0;
	    pua43_n_sumsq=0.0;
	    noisefraction=1.0-pua43_noisedecay;
	    for (i=0; i<43; i++)
		{
		/* RC decayed sum of noise for each of 43 freq:  */
		if (bits_8_16)
		  {
		  pua43_n[i]=noisefraction*(pd16[b0[fft_bw]-2
			+4*i]+pd16[b0[fft_bw]+2+4*i])
			+pua43_noisedecay*pua43_n[i];
		  }
		else
		  {
	    	  pua43_n[i]=noisefraction*
			(power10[raw_data[b0[fft_bw]-2+4*i]]
			+power10[raw_data[b0[fft_bw]+2+4*i]])+
			+pua43_noisedecay*pua43_n[i];
		  }
		pua43_n_sum += pua43_n[i];
		pua43_n_sumsq += (pua43_n[i]*pua43_n[i]);
		}
	    pua43_n_sum /= 43.0;

	    }    /* if what==3 */
	else if (what==4)		/* Update status box for transmit*/
	    {
	    if (current_info_box == 0)  return;
	    lb_open(&lb_pua43, 0);		/* Draw the box */

	    if (pua14_28)
	        lb_line(&lb_pua43, 1, BOLD, " 28 Char total", 0);
	    else
	        lb_line(&lb_pua43, 1, BOLD, " 14 Char total", 0);

	    lb_line(&lb_pua43, 2, BOLD, " ", 0);
	    lb_line(&lb_pua43, 3, BOLD, " ", 0);
	    lb_line(&lb_pua43, 4, BOLD, " ", 0);
	    lb_line(&lb_pua43, 5, BOLD, " ", 0);
	    lb_line(&lb_pua43, 6, BOLD, " ", 0);
	    lb_line(&lb_pua43, 7, BOLD, " ", 0);
	    lb_line(&lb_pua43, 8, BOLD, " ", 0);
	    lb_line(&lb_pua43, 9, BOLD, " ", 0);
	    lb_line(&lb_pua43, 11, BOLD, " ", 0);

	    if (pua43_dorandom)
		sprintf(cc+250, "Random stir=%d", pua43_rnd[code_min] );
	    else
		sprintf(cc+250, "Not Randomized");
	    lb_line(&lb_pua43, 10, BOLD, cc+250, 0);

	    savch=pua43_xstring[14];
	    pua43_xstring[14]=0;
	    sprintf(cc+325, "|%s|", pua43_xstring);
	    lb_line(&lb_pua43, 13, BOLD, cc+325, 0);

	    pua43_xstring[14]=savch;
	    if (pua14_28)
		sprintf(cc+350, "|%s|", &pua43_xstring[14]);
	    else
		sprintf(cc+350, " ");
	    lb_line(&lb_pua43, 14, BOLD, cc+350, 0);

	    i=time_tag/2;
	    if (pua14_28==FALSE && i>=14)  i -= 14;
	    sprintf(cc+300, "t=%d sec;  Char=%c", time_tag, pua43_xstring[i]);
	    lb_line(&lb_pua43, 12, BOLD, cc+300, 0);

	    if (current_info_box==&lb_pua43)
	       lb_update(&lb_pua43);	/* Redraw the info box with new data */
	    }
	}

/* cnvt43 - Converts a (0,42) tone position to an ASCII value.
 */
char cnvt43(char c43)
	{
	if (c43==10) c43=-2;		/* Period */
	else if (c43==11) c43=-4;	/* Comma */
	else if (c43==12) c43=-16;	/* Space */
	else if (c43==13) c43=-1;	/* Forward Slash */
	else if (c43==14) c43=-13;	/* # = Received Msg */
	/* Note 15, the '?', stays put */
	else if (c43==16) c43=-12;	/* $ = Shift next char */
	else if (c43==31) c43=63;	/* Convert O to o (confuse with 0) */
	return 48+c43;
	}

/*  EME2_UPDATE()  -  Accumulates data from 2 second EME pulses.
 *  Controls status box, but does not control activities, except to zero sums.
 *  what:   -1  Reset and zero sums.
 *           0  Close status box, but leave sums as they are.
 *           1  Open status box, use sums as they are.
 *           2  Update all displays.
 */
void eme2_update(int what)
	{
	char cc[600];
	int i, j, i2;
	double ep1, ep2;		/* Working variables */
	/* The global eme_pwr_sum[] is the accumulated power at each of the
	 * 21 frequencies.  10 noise freq below, 10 noise freq above
	 * and 1 signal freq at center. Transmit and receive set to make output
	 * always peak at
	 * fft_bw==2  (4800 Hz span)  900 Hz (bin 96)
	 * fft_bw==1  (2400 Hz span)  600 Hz (bin 128)
	 * fft_bw==0  (1200 Hz span)  323 Hz (bin 138)
	 *
	 *   FFT  low disp  low detail  center  high detail  high disp
	 *  1200     90        128       138        148         186
	 *  2400     80        118       128        138         176
	 *  4800     48         86        96        106         144
	 */
	static double    pwr_sum21=1.0e-200;	/* Power sum across all 21 */
	double           s2ndb;
	char             ch_record;

	if (what==-1)			/* Reset by zeroing power sums */
	   {for (i=0; i<256; i++)
	      {
	      lt_pwrdat[i]=0.0;
	      /* Also clear the normalization for line straightening */
	      lt_dnorm[i]=0.0;
	      }
	    for(i=0; i<21; i++)  eme_pwr_sum[i]=0.0;
	    pwr_sum21=1.0e-200;
	    return;
	   }
	else if (what==0)		/* Close EME2 status box, no reset */
	   {
	   lb_close(&lb_eme2, 0);		/* Undisplay the box */
	   current_info_box=0;		/* Tell everybody */
	   }
	else if (what==1)		/* Open EME2 data box */
	   {
	    if (current_info_box  &&  (current_info_box != &lb_eme2))
		lb_close(current_info_box, 0);
	    lb_create(&lb_eme2, "EME-2 SEC",
		 450, 148, 638, 380, 0, NON_MODAL);
	   }
	else if (what==2)		/* Update power sums and update disp */
	   {
	    lb_open(&lb_eme2, 0);		/* Draw the box */

	    /* Fixed text in EME2 box:     */
	    lb_line(&lb_eme2, 19, TXT, "Close=Alt A    Clear=^W", 0);
	    lb_line(&lb_eme2, 20, TXT, "Straighten=^E          ", 0);
	    lb_line(&lb_eme2, 21, TXT, "Changes=Alt B          ", 0);

/* Switch to lt_pwr... and will allow for 8/16 bit data, don't need
 * to clear two different items,...  
 */

	    pwr_sum21=1.0e-200;
	    for (i=0; i<21; i++)		/* Cumulate power */
		/* lt_dnorm[i] is dB noise power at freq i (0, 255) */
		{eme_pwr_sum[i]=upow10(-lt_dnorm[lt_i2+i])*lt_pwrdat[lt_i2+i];
		 if (i!=9 && i!=10 && i!=11)
		     pwr_sum21 += eme_pwr_sum[i];
		}
	    /* There are 18 noise points and only 1 sig, thus 12.5527 dB */
	    s2ndb=12.5527+p2db(eme_pwr_sum[10]/pwr_sum21);

	    /* Find noise power, assuming -174 dBm/Hz noise.
	     * 3.98107e-18 is KTo and the bandwidth is set by the FFT
	     * bid width along with the broadening of the windowing fcn.
	     */
	    ep1=3.98107e-18*bin_width[fft_bw]*noise_bw[do_window];

	    /* -174 dBm is for 290 deg K. Correct for actual noise temp.
	     */
	    ep1 *= (0.003448*eme2_te);		/* 0.003448=1/290 */

	    /* Update the EME2 data box */
	    if (eme_record) ch_record='Y';  else ch_record='N';
	    sprintf(cc+25, " Points=%6ld   Rcd=%c",
			eme_pulse_count, ch_record);
	    lb_line(&lb_eme2, 1, BOLD, cc+25, 0);

	    sprintf(cc+50, " NonCohInt=%4.1lfdB",
			0.5*p2db((double)eme_pulse_count));
	    lb_line(&lb_eme2, 2, BOLD, cc+50, 0);

	    sprintf(cc+75, " Te (.cfg) = %d K", (int)(0.5+eme2_te));
	    lb_line(&lb_eme2, 3, BOLD, cc+75, 0);

	    if (fft_bw==0)
		sprintf(cc+100, "  323Hz S+N/N=%.3lfdB", s2ndb);
	    else if (fft_bw==1)
		sprintf(cc+100, "  600Hz S+N/N=%.3lfdB", s2ndb);
	    else
		sprintf(cc+100, "  900Hz S+N/N=%.3lfdB", s2ndb);

	    lb_line(&lb_eme2, 4, BOLD, cc+100, 0);

	    ep2=upow10(0.1*s2ndb)-1.0;
	    if (ep2>0.0)
		sprintf(cc+125, "  Cntr Sig=%6.1lfdBm", p2db(ep2*ep1) );
	    else
		sprintf(cc+125, "  Cntr Sig= ---- dBm");
	    lb_line(&lb_eme2, 5, BOLD, cc+125, 0);

	    for (i=0; i<9; i++)
		{
		 sprintf(cc+175+i*25, " %7.3lf   N%d  %7.3lf",
		    12.5527+p2db(eme_pwr_sum[ 9-i]/pwr_sum21), i+1,
		    12.5527+p2db(eme_pwr_sum[11+i]/pwr_sum21) );
		 lb_line(&lb_eme2, 7+i, BOLD, cc+175+i*25, 0);
		}
		 /* Due to the extra char, the last line is different: */
		 sprintf(cc+400, " %7.3lf   N10 %7.3lf",
		    12.5527+p2db(eme_pwr_sum[ 9-i]/pwr_sum21),
		    12.5527+p2db(eme_pwr_sum[11+i]/pwr_sum21) );

		 lb_line(&lb_eme2, 16, BOLD, cc+400, 0);

		 if (eme2_format && 1)
		     sprintf(cc+425, " Dop Cor, Ran=%d Hz", (int)random_freqL);
		 else
		     sprintf(cc+425, " No Dop,  Ran=%d Hz", (int)random_freqL);
		 lb_line(&lb_eme2, 17, BOLD, cc+425, 0);

		 sprintf(cc+450, " Pav=%6.2lf  P=%6.2lf",
		     p2db(paveave), p2db(pave));
		 lb_line(&lb_eme2, 18, BOLD, cc+450, 0);

	    lb_update(&lb_eme2);	/* Redraw the info box with new data */
	   }
	else return;
	}

/*  LTERM_INT_UPDATE()  -  Accumulates data from LTI Long Term Integration.
 *  Controls status box, but does not control activities, except to zero sums.
 *  what:   -1  Reset and zero sums.
 *           0  Close status box, but leave sums as they are.
 *           1  Open status box, use sums as they are.
 *           2  Update all displays.
 *
 * This is basically the same as the EME2 receive function, and is never used
 * at the same time. Therefore, the same variables, such as lt_pwrdat[] are
 * reused here.
 */
void lterm_int_update(int what)
	{
	static char cc[600];
	int i, j, i2;
	double ep1, ep2;		/* Working variables */
	static double    pwr_sum21=1.0e-200;	/* Power sum across all 21 */
	double           s2ndb;
	/* The global eme_pwr_sum[] is the accumulated power at each of the
	 * 21 frequencies.  10 noise freq below, 10 noise freq above
	 * and 1 signal freq at center.
	 */

	if (what==-1)			/* Reset by zeroing power sums */
	   {for (i=0; i<256; i++)
	      {
	      lt_pwrdat[i]=0.0;
	      /* Also clear the normalization for line straightening */
	      lt_dnorm[i]=0.0;
	      }
	    for(i=0; i<21; i++)  eme_pwr_sum[i]=0.0;
	    pwr_sum21=1.0e-200;
	    return;
	   }
	else if (what==0)		/* Close LTERM INT status box */
	   {
	   lb_close(&lb_lterm_int, 0);	/* Undisplay the box */
	   current_info_box=0;		/* Tell everybody */
	   }
	else if (what==1)		/* Open LTERM_INT data box */
	   {
	    if (current_info_box  &&  (current_info_box != &lb_lterm_int))
		lb_close(current_info_box, 0);
	    lb_create(&lb_lterm_int, "LTI LONG TERM TRACE",
		 450, 148, 638, 380, 0, NON_MODAL);
	   }
	else if (what==2)		/* Update power sums and update disp */
	   {
	    lb_open(&lb_lterm_int, 0);		/* Draw the box */

	    /* Fixed text in LTERM_INT box:     */
	    lb_line(&lb_lterm_int, 19, TXT, "Close=Alt A    Clear=^W", 0);
	    lb_line(&lb_lterm_int, 20, TXT, "Straighten=^E          ", 0);
	    lb_line(&lb_lterm_int, 21, TXT, "Changes=Alt B          ", 0);
/* Switch to lt_pwr... and will allow for 8/16 bit data, don't need
 * to clear two different items,...  
 */

	    /* lt_i2 is a global variable that is the first bin of the 21 for
	       detailed data */
	    pwr_sum21=1.0e-200;
	    for (i=0; i<21; i++)		/* Cumulate power */
		/* lt_dnorm[i] is dB noise power at freq i (0, 255) */
		{eme_pwr_sum[i]=upow10(-lt_dnorm[lt_i2+i])*lt_pwrdat[lt_i2+i];
		 if (i!=9 && i!=10 && i!=11)
		     pwr_sum21 += eme_pwr_sum[i];
		}
	    /* There are 18 noise points and only 1 sig, thus 12.5527 dB */
	    s2ndb=12.5527+p2db(eme_pwr_sum[10]/pwr_sum21);

	    /* Find noise power, assuming -174 dBm/Hz noise.
	     * 3.98107e-18 is KTo and the bandwidth is set by the FFT
	     * bid width along with the broadening of the windowing fcn.
	     */
	    ep1=3.98107e-18*bin_width[fft_bw]*noise_bw[do_window];

	    /* -174 dBm corresponds to 290 deg K.
	     * Correct here for actual noise temp */
	    ep1 *= (0.003448*eme2_te);		/* 0.003448=1/290 */

	    /* Update the LTI data box */
	    sprintf(cc, " Pts=%ld  %4.1lf dB", eme_pulse_count,
			0.5*p2db((double)eme_pulse_count));
	    lb_line(&lb_lterm_int, 0, BOLD, cc, 0);

	    sprintf(cc+25, " Te (.cfg) = %d K", (int)(0.5+eme2_te));
	    lb_line(&lb_lterm_int, 1, BOLD, cc+25, 0);

	    sprintf(cc+50, " f1=%d fc=%d f5=%d",
		(int)(0.5+df[fft_bw]*(double)lt_i1),
		(int)(0.5+df[fft_bw]*(double)lt_i3),
		(int)(0.5+df[fft_bw]*(double)lt_i5));
	    lb_line(&lb_lterm_int, 2, BOLD, cc+50, 0);

	    ep2=upow10(0.1*s2ndb)-1.0;
	    if (ep2>0.0)
		sprintf(cc+75, "  Cntr Sig=%6.1lfdBm", p2db(ep2*ep1) );
	    else
		sprintf(cc+75, "  Cntr Sig= ---- dBm");
	    lb_line(&lb_lterm_int, 3, BOLD, cc+75, 0);

	    lb_line(&lb_lterm_int, 4, BOLD, " ", 0);
	    lb_line(&lb_lterm_int, 5, BOLD, " -----  (S+N)/N  -----", 0);

	    sprintf(cc+150, "      N0=%.3lfdB", s2ndb);
	    lb_line(&lb_lterm_int, 6, BOLD, cc+150, 0);

	    for (i=0; i<9; i++)
		{
		 sprintf(cc+175+i*25, " %7.3lf   N%d  %7.3lf",
		    12.5527+p2db(eme_pwr_sum[ 9-i]/pwr_sum21), i+1,
		    12.5527+p2db(eme_pwr_sum[11+i]/pwr_sum21) );
		 lb_line(&lb_lterm_int, 7+i, BOLD, cc+175+i*25, 0);
		}
	    /* Due to the extra char, the last line is different: */
	    sprintf(cc+400, " %7.3lf   N10 %7.3lf",
	       12.5527+p2db(eme_pwr_sum[ 9-i]/pwr_sum21),
	       12.5527+p2db(eme_pwr_sum[11+i]/pwr_sum21) );
	    lb_line(&lb_lterm_int, 16, BOLD, cc+400, 0);

	    sprintf(cc+425, " Random stir=%d Hz", (int)(0.5+lti_frandom));
	    lb_line(&lb_lterm_int, 17, BOLD, cc+425, 0);

	    sprintf(cc+450, " Pav=%6.2lf  P=%6.2lf",
		p2db(paveave), p2db(pave));
	    lb_line(&lb_lterm_int, 18, BOLD, cc+450, 0);

	    lb_update(&lb_lterm_int);	/* Redraw the info box with new data */
	   }
	else return;
	}

static void send(int cw)
	{
	if (cw)
	    {
	    ASYNCOUT(ESC_CHAR);  ASYNCOUT('T');
	    ASYNCOUT(17); ASYNCOUT(2);
	    ASYNCOUT(fgainl[xmit_power]); ASYNCOUT(fgainh[xmit_power]);
	    }
	else
	    {
	    ASYNCOUT(ESC_CHAR);  ASYNCOUT('T');
	    ASYNCOUT(1); ASYNCOUT(2);
	    ASYNCOUT(fgainl[xmit_power]); ASYNCOUT(fgainh[xmit_power]);
	    }
	}

/* lhl_code returns fsk codes as suggested by W7LHL.  Based on Morse code
 * each letter is sent as a a group of dot and dash frequencies
 * followed, possibly, by a multiplier if needed.  The same
 * multipliers apply to either dots or to dashes.  The offset
 * frequencies are:
 *		Separator	 600 Hertz	600+9.375*0
 *		Dot		 750		600+9.375*16
 *		Dash		 900		600+9.375*32
 *		x2		1050		600+9.375*48
 *		x3		1200		600+9.375*64
 *		x4		1350		600+9.375*80
 *		x5		1500		600+9.375*96
 *
 * This function returns 0 if unsuccessful, due to an illegal character.
 * Otherwise, returns the number of tones to be sent.
 *
 * The structure lhl is:
 *    lhl.cc       unsigned char   The character being sent
 *    lhl.ntone    int             Number of tones in code
 *    lhl.curtone  int             The current tone index being sent
 *    lhl.tone[11] int             1 to 12 tone frequencies, in sequence
 */
int lhl_code(unsigned char c, struct lhl *lcode)
	{
	int j, k, ij;
	long int ic, jtest;

	lcode->cc=c;
        lcode->curtone=0;

	if (c=='+')                     /* Special case of test pattern */
	    {lcode->ntone=7;
	     lcode->tone[0]=600;
	     lcode->tone[1]=750;
	     lcode->tone[2]=900;
	     lcode->tone[3]=1050;
	     lcode->tone[4]=1200;
	     lcode->tone[5]=1350;
	     lcode->tone[6]=1500;
             return 7;
            }
	if (c=='&')			/* Special case of CW ID */
	    {lcode->ntone=1;
	     lcode->tone[0]=LHL_CWID;
	     return 1;
	    }

	for(k=0; k<11; k++)  lcode->tone[k]=0;

	if (c>95)  c -= 64;		/* Convert lc to caps */
	else c -= 32;			/* Move subscript to (0, 63) */
        if (c>63) return 0;             /* Outside table range */
	ic = m[(int)c];			/* Ch to  code lookup */
        if (ic==0) return 0;		/* Not in table */

	k=0;
	jtest=1000000000;
	for(j=0; j<10; j++)		/* First move past leading zeros */
	    {if ((ic/jtest) != 0) break;
             jtest /= 10;
	    }
	j++;				/* The first 1 of ic is start  */
	ic -= jtest;			/*    symbol and we skip over it */
	jtest /= 10;

	for(; j<10; j++)		/* Now fill out code, j comes set */
	    {ij=(int)(ic/jtest);
	     lcode->tone[k++]=750+150*ij;
	     ic -= jtest*(long int)ij;
             jtest /= 10;
	    }
	lcode->tone[k]=LHL_SEP;
	return k+1;
	}

/* pua43_freq(c) returns the fsk tone (offset above the listed transmit
 * frequency for any of the 43 characters:
 * 	0 to 9
 *	period, comma, question mark, forward slash
 *	space, message received, shift
 *	A to Z
 * These are sent on offset frequencies 450 + 5*i*df[fft_bw]
 * where df is the bin size in Hz.
 *
 * If the variable pua43_dorandom is TRUE, there is a rotation added to
 * the table of -pua43_rnd[code_min].  This randomizes the transmit freq
 * used for the 43 characters, with wrap-around if the range is exceeded.
 *
 * The frequency offset is set by table x43[i].
 */
double pua43_freq(unsigned char c)
	{
	int cint;

	if      (c==32) c=12;		/* Space */
	else if (c==46) c=10;		/* Period */
	else if (c==44) c=11;		/* Comma */
	else if (c==47) c=13;		/* Forward Slash */
	else if (c==35) c=14;		/* # for 'Received' */
	else if (c==63) c=15;		/* Question mark */
	else if (c==36) c=16;		/* $ for 'Shift' */
	else if (c>=48 && c<=57) c-=48; /* Numbers 0 to 9 */
	else if (c>=65 && c<=90) c-=48; /* Letters A to Z */
	else if (c>=97 && c<=122) c-=80; /* Letters a to z */
	else c=15;			/* Everything else to ? */

	if (pua43_dorandom)		/* Randomize frequencies? */
	    {
	    cint=(int)c-pua43_rnd[code_min];	/* Yes */
	    if (cint<0)  cint += 43;		/* Adjust to (0, 42) */
	    }

	/* The following finds the actual offset frequency for transmit */
	return df[fft_bw]*(double)(b0[fft_bw]+4*cint);
	}

/* lti_freq() returns the fsk tone (offset above the listed transmit
 * frequency for current time and in LTI mode. The center frequency is set by
 * lterm_int_center, converted to a bin number as lt_lt_i3. The randomization
 * comes from adding on bins equal to the value from pua43_rnd[]-21, all
 * multiplied by a spreading constant of (0, 4) where the value 0 is no
 * randomization. This process keeps the center frequency centered on
 * an FFT bin.
 * Also sets global double lti_frandom to random offset value.
 */
double lti_freq(void)
	{
	lti_frandom
	    =0.0238095238*lti_freq_range*(double)(pua43_rnd[code_min]-21);
	return lti_frandom + (double)lt_i3*df[fft_bw];
	}

/* A modal dialog box for EME-2 mode.
 */
void eme2_modalbox(void)
	{
	static struct edit_info   el1, el3, el4;
	static struct radio_info  rl2;
	static struct check_info ck1;
	double nbdb;
	int bpush;
	char eline1[12], eline2[12], eline3[12], eline4[12];
	unsigned int edresult, rbui, ckui;

	lb_create(&lb1, "EME2 ECHO MODE", 160, 175, 370, 440, 0, MODAL);
	lb_open(&lb1, 1);

	/* Fixed text in modal box:     */
	lb_line(&lb1, 21, BOLD, "\030 and \031 Move to Next Item", 0);
	lb_line(&lb1, 22, BOLD, "SPACE=Change Button/Box", 0);
	lb_line(&lb1, 23, BOLD, "TEXT: Backspace & Type", 0);
	lb_line(&lb1, 24, BOLD, "ENTER=Close Dialog", 0);

	/* Editable text areas:         */

	/* The field to be edited:    */
	sprintf(eline1, "%.3d", (int)(eme_freq_range+0.5));
	lb_create_el(&el1, 5, 151, &eline1[0],
		ED_INACTIVE, 0,	strlen(eline1) );
	lb_line(&lb1, 3, EDIT, " Random Spread, Hz", &el1);

	strcpy(eline3, code_id);	/* This is the line to be edited */
	lb_create_el(&el3, 10, 111, &eline3[0],
		ED_INACTIVE, 0, strlen(eline3));
	lb_line(&lb1, 16, EDIT, "Ident String", &el3);

	if (code_nb_ratio<=0.0)  code_nb_ratio=0.10223294e-10;
	if (code_nb_ratio>=9.7723e9) code_nb_ratio=9.7723e9;
	sprintf(eline4, "%.2lf", 10.0*log10(code_nb_ratio));
	lb_create_el(&el4, 6, 143, &eline4[0], ED_INACTIVE, 0, strlen(eline4));
	lb_line(&lb1, 18, EDIT, "Noise Blank, dB", &el4);

	/* Define the max dB/div Radio Button family:     */
	if (min_lt_dd<=1.600001)
	    {
	    min_lt_dd=1.6;
	    bpush=9;
	    }
	else if (min_lt_dd<=3.200001)
	    {
	    min_lt_dd=3.2;
	    bpush=10;
	    }
	else if (min_lt_dd<=8.000001)
	    {
	    min_lt_dd=8.0;
	    bpush=11;
	    }
	else
	    {
	    min_lt_dd=16.0;
	    bpush=12;
	    }

	lb_create_rb(&rl2, 0, bpush, 20);	/* Note: bpush pushed */
	/* Each line is defined separately with lb_line(..):             */
	lb_line(&lb1,  9, RADIO, "LTerm Trace max 10 dB/", &rl2);
	lb_line(&lb1, 10, RADIO, "LTerm Trace max  5 dB/", &rl2);
	lb_line(&lb1, 11, RADIO, "LTerm Trace max  2 dB/", &rl2);
	lb_line(&lb1, 12, RADIO, "LTerm Trace max  1 dB/", &rl2);

	/* Record check box:     */
	if (eme_record)			/* Show box checked */
	    lb_create_check(&ck1, 0, TRUE, 20);
	else				/* Show box unchecked */
	    lb_create_check(&ck1, 0, FALSE, 20);
	lb_line(&lb1, 7, CHECK, "Record Data to Disk", &ck1);

	lb_update(&lb1);

 e3:	/* Now edit the active, cursor on, field:  */
	edresult=lb_edit_line(&lb1, &el1);
	if (edresult==CUR_UP)   goto e18;
	else if (edresult==ENTER) goto lb1end;

 c7:	ckui=lb_check_mark(&lb1, &ck1);
	if     (ckui==CUR_UP)  goto e3;
	else if(ckui==ENTER)     goto lb1end;

 r9:	rbui=lb_radio_button(&lb1, 9, &rl2);
	if     (rbui==CUR_UP)  goto c7;
	else if(rbui==ENTER)   goto lb1end;

 r10:	rbui=lb_radio_button(&lb1, 10, &rl2);
	if     (rbui==CUR_UP)  goto r9;
	else if(rbui==ENTER)   goto lb1end;

 r11:	rbui=lb_radio_button(&lb1, 11, &rl2);
	if     (rbui==CUR_UP)  goto r10;
	else if(rbui==ENTER)   goto lb1end;

 r12:	rbui=lb_radio_button(&lb1, 12, &rl2);
	if     (rbui==CUR_UP)  goto r11;
	else if(rbui==ENTER)   goto lb1end;

 e16:	edresult=lb_edit_line(&lb1, &el3);
	if (edresult==CUR_UP)   goto r12;
	else if (edresult==ENTER) goto lb1end;

 e18:	edresult=lb_edit_line(&lb1, &el4);
	if (edresult==CUR_UP)   goto e16;
	else if (edresult==CUR_DOWN) goto e3;
	else if (edresult==ENTER) goto lb1end;

 lb1end:/* As we exit, the variables must be changed to agree with the
	 * indicated states of the buttons and edit fields:
	 */
	eme_freq_range=atof(eline1);
	if (eme_freq_range < 0.0)
	    {
	    eme_freq_range=0.0;
	    lb_msg(170, 260, "WARNING",
		"Random Spread must be positive",
		"      Value set to zero");
	    }
	if (eme_freq_range >2000.0)
	    {
	    eme_freq_range=2000.0;
	    lb_msg(170, 260, "WARNING",
		" Random Spread max of 2000 Hz",
		"    Value set to 2000 Hz");
	    }

	if      ((1&rl2.state[ 9])==1) min_lt_dd= 1.6;
	else if ((1&rl2.state[10])==1) min_lt_dd= 3.2;
	else if ((1&rl2.state[11])==1) min_lt_dd= 8.0;
	else if ((1&rl2.state[12])==1) min_lt_dd=16.0;

	delay(500);
	/* Assign variables: */
	strcpy(code_id, eline3);

	nbdb=atof(eline4);
	if (nbdb < -99.9)
	    {
	    nbdb=-99.9;
	    lb_msg(170, 260, "WARNING",
		"  Noise Blank must be >-99.9",
		"      Value set to -99.9");
	    }
	if (nbdb > 99.9)
	    {
	    nbdb=99.9;
	    lb_msg(170, 260, "WARNING",
		"  Noise Blank max of 99.9 dB",
		"      Value set to 99.9");
	    }
	code_nb_ratio=upow10(0.1*nbdb);

	if (ck1.state & CHECK_MARK)
	    eme_record=TRUE;
	else
	    eme_record=FALSE;
	set_bins();			/* Update variables */
	lb_close(&lb1, 1);
	}

/*************************************************************************
 * A modal dialog box for LHL-7 mode.
 *
 */
void lhl7_modalbox(void)
	{
	static struct edit_info   el1;
	static struct radio_info  rl1, rl2, rl3, rl4;
	char eline1[12];
	unsigned int edresult, rbui;
	int binline, paline, outline, nx2line;

	lb_create(&lb2, "LHL-7 SETUP", 160, 175, 370, 440, 0, MODAL);
	lb_open(&lb2, 1);

	/* Fixed text in modal box:     */
	lb_line(&lb2, 22, BOLD, "SPACE changes buttons", 0);
	lb_line(&lb2, 23, BOLD, "\030 and \031 Move to Next Item", 0);
	lb_line(&lb2, 24, BOLD, "ENTER=Close    ESC=Cancel", 0);

	lb_line(&lb2, 15, BOLD, "BACKSPACE and type:", 0);
	lb_create_el(&el1, 10, 111, &eline1[0], ED_INACTIVE, 0, 8);
	strcpy(eline1, code_id);	/* This is the line to be edited */
	lb_line(&lb2, 16, EDIT, "Ident String", &el1);

	/* Define the Radio Button families:     */
	if (lhl_bins >= 9)    binline=5;
	else if (lhl_bins==7) binline=4;
	else if (lhl_bins==5) binline=3;
	else if (lhl_bins==3) binline=2;
	else                  binline=1;
	lb_create_rb(&rl1, 0, binline, 20);	/* Note: binline pushed */

	/* Each line is defined separately with lb_line(..):             */
	lb_line(&lb2, 1, RADIO, "Center bin only", &rl1);
	lb_line(&lb2, 2, RADIO, "3 Bins for signals", &rl1);
	lb_line(&lb2, 3, RADIO, "5 Bins for signals", &rl1);
	lb_line(&lb2, 4, RADIO, "7 Bins for signals", &rl1);
	lb_line(&lb2, 5, RADIO, "9 Bins for signals", &rl1);

	/* Peak/Ave family:     */
	if (lhl_peak_ave) paline=8; else paline=7;
	lb_create_rb(&rl2, 0, paline, 20);	/* Note: lhl_peak_ave pushed */
	lb_line(&lb2, 7, RADIO, "Peak Level in Bins", &rl2);
	lb_line(&lb2, 8, RADIO, "Average Level in Bins", &rl2);

	/* Define the code_nx2 Radio Button family:     */
	if (code_nx2 >= 30)    nx2line=14;
	else if (code_nx2==10) nx2line=13;
	else if (code_nx2==5)  nx2line=12;
	else if (code_nx2==2)  nx2line=11;
	else                   nx2line=10;
	lb_create_rb(&rl3, 0, nx2line, 20);	/* Note: nx2line pushed */

	lb_line(&lb2, 10, RADIO, "2  Seconds per tone", &rl3);
	lb_line(&lb2, 11, RADIO, "4  Seconds per tone", &rl3);
	lb_line(&lb2, 12, RADIO, "10 Seconds per tone", &rl3);
	lb_line(&lb2, 13, RADIO, "20 Seconds per tone", &rl3);
	lb_line(&lb2, 14, RADIO, "60 Seconds per tone", &rl3);

	/* 1/2 Line family:     */
	if (lhl7_2line) outline=20; else outline=19;
	lb_create_rb(&rl4, 0, outline, 20);	/* Note: lhl_peak_ave pushed */
	lb_line(&lb2, 19, RADIO, "Single Line Output", &rl4);
	lb_line(&lb2, 20, RADIO, "2 Line Output", &rl4);

	lb_radio_active(&rl1, 1, RADIO_ACTIVE); /* Start point in lb2 */

	lb_update(&lb2);

 r1:	/* Now give user opportunity to push button 1: */
	rbui=lb_radio_button(&lb2, 1, &rl1);
	/* Find out how exit was done and react:      */
	if     (rbui==CUR_UP)  goto r20;
	else if(rbui==ENTER)   goto lb2end;

 r2:	rbui=lb_radio_button(&lb2, 2, &rl1);
	if     (rbui==CUR_UP)  goto r1;
	else if(rbui==ENTER)   goto lb2end;

 r3:	rbui=lb_radio_button(&lb2, 3, &rl1);
	if     (rbui==CUR_UP)  goto r2;
	else if(rbui==ENTER)   goto lb2end;

 r4:	rbui=lb_radio_button(&lb2, 4, &rl1);
	if     (rbui==CUR_UP)  goto r3;
	else if(rbui==ENTER)   goto lb2end;

 r5:	rbui=lb_radio_button(&lb2, 5, &rl1);
	if     (rbui==CUR_UP)  goto r4;
	else if(rbui==ENTER)   goto lb2end;

 r7:	rbui=lb_radio_button(&lb2, 7, &rl2);
	if     (rbui==CUR_UP)  goto r5;
	else if(rbui==ENTER)     goto lb2end;

 r8:	rbui=lb_radio_button(&lb2, 8, &rl2);
	if     (rbui==CUR_UP)  goto r7;
	else if(rbui==ENTER)     goto lb2end;

 r10:	rbui=lb_radio_button(&lb2, 10, &rl3);
	if     (rbui==CUR_UP)  goto r8;
	else if(rbui==ENTER)     goto lb2end;

 r11:	rbui=lb_radio_button(&lb2, 11, &rl3);
	if     (rbui==CUR_UP)  goto r10;
	else if(rbui==ENTER)     goto lb2end;

 r12:	rbui=lb_radio_button(&lb2, 12, &rl3);
	if     (rbui==CUR_UP)  goto r11;
	else if(rbui==ENTER)     goto lb2end;

 r13:	rbui=lb_radio_button(&lb2, 13, &rl3);
	if     (rbui==CUR_UP)  goto r12;
	else if(rbui==ENTER)     goto lb2end;

 r14:	rbui=lb_radio_button(&lb2, 14, &rl3);
	if     (rbui==CUR_UP)  goto r13;
	else if(rbui==ENTER)     goto lb2end;

 e16:	edresult=lb_edit_line(&lb2, &el1);
	if (edresult==CUR_UP)   goto r14;
	else if (edresult==ENTER) goto lb2end;

 r19:	rbui=lb_radio_button(&lb2, 19, &rl4);
	if     (rbui==CUR_UP)  goto e16;
	else if(rbui==ENTER)     goto lb2end;

 r20:	rbui=lb_radio_button(&lb2, 20, &rl4);
	if     (rbui==CUR_UP)  goto r19;
	else if(rbui==ENTER)     goto lb2end;
	goto r1;

 lb2end: delay(200);
	/* As we exit, the variables must be changed to agree with the
	 * indicated states of the buttons and edit fields:
	 */
	if ((1&rl1.state[1])==1)      lhl_bins=1;
	else if ((1&rl1.state[2])==1) lhl_bins=3;
	else if ((1&rl1.state[3])==1) lhl_bins=5;
	else if ((1&rl1.state[4])==1) lhl_bins=7;
	else if ((1&rl1.state[5])==1) lhl_bins=9;

	lhl_peak_ave=(1&rl2.state[8]);

	if ((1&rl3.state[10])==1)      code_nx2=1;
	else if ((1&rl3.state[11])==1) code_nx2=2;
	else if ((1&rl3.state[12])==1) code_nx2=5;
	else if ((1&rl3.state[13])==1) code_nx2=10;
	else if ((1&rl3.state[14])==1) code_nx2=30;
	set_code_nx2();			/* Display nx2 and adjust spec_ave */

	lhl7_2line=(1&rl4.state[20]);

	strcpy(code_id, eline1);	/* Get any changes to code_id */
	lb_close(&lb2, 1);
	}

/*************************************************************************
 * A modal dialog box for PUA43 mode.
 * The box structure is lb3.
 */
void pua43_modalbox(void)
	{
	static struct edit_info   el1, el2, el3, el4, el5, el6;
	static struct radio_info  rl1;
	char eline1[12], eline2[15], eline3[15], eline4[12],
		 eline5[12], eline6[12];
	unsigned int edresult, rbui;
	int rline;
	int sl, i;
	double nbdb;

	lb_create(&lb3, "PUA43 SETUP", 160, 175, 370, 440, 0, MODAL);
	lb_open(&lb3, 1);

	/* Fixed text in modal box:     */
	lb_line(&lb3, 21, BOLD, "TEXT: Backspace & Type", 0);
	lb_line(&lb3, 22, BOLD, "SPACE changes buttons", 0);
	lb_line(&lb3, 23, BOLD, "\030 and \031 Move to Next Item", 0);
	lb_line(&lb3, 24, BOLD, "ENTER=Close    ESC=Cancel", 0);

	sl=strlen(code_id);
	if (sl>9)
	    {
	    sl=9;
	    code_id[9]=0;
	    }

	lb_create_el(&el1, 10, 111, &eline1[0], ED_INACTIVE, 0, sl);
	strcpy(eline1, code_id);	/* This is the line to be edited */
	lb_line(&lb3, 11, EDIT, "Ident String", &el1);

	lb_line(&lb3, 14, BOLD, "Transmit Lines:", 0);
	sl=strlen(pua43_xstring);
	if (sl>14)
	    {
	    for (i=0; i<14; i++)
		eline2[i]=pua43_xstring[i];
	    eline2[14]=0;
	    sl=14;
	    }
	else
	    strcpy(eline2, pua43_xstring);
	lb_create_el(&el2, 15, 71, &eline2[0], ED_INACTIVE, 0, sl);
	lb_line(&lb3, 15, EDIT, "First 14:", &el2);

	sl=strlen(pua43_xstring);
	if (sl<=14)			/* There is no line 2 */
	    {
	    sl=0;
	    eline3[0]=0;
	    }
	else
	    {
	    if (sl>28) sl=28;		/* Whoops, too long */
	    for (i=0; i<(sl-14); i++)
		eline3[i]=pua43_xstring[i+14];
	    eline3[sl-14]=0;
	    sl -= 14;
	    }
	lb_create_el(&el3, 15, 71, &eline3[0], ED_INACTIVE, 0, sl);
	lb_line(&lb3, 17, EDIT, "Scnd 14:", &el3);

	if (code_nb_ratio<=0.0)  code_nb_ratio=0.10223294e-10;
	if (code_nb_ratio>=9.7723e9) code_nb_ratio=9.7723e9;
	sprintf(eline6, "%.2lf", 10.0*log10(code_nb_ratio));
	lb_create_el(&el6, 6, 143, &eline6[0], ED_INACTIVE, 0, strlen(eline6));
	lb_line(&lb3, 19, EDIT, "Noise Blank, dB", &el6);

	/* An edit field for the noise decay rate: */
	sprintf(eline4,"%.6f", pua43_noisedecay); /* The field to be edited */
	sl=strlen(eline4);
	lb_create_el(&el4, 9, 119, &eline4[0], ED_INACTIVE, 0, sl);
	lb_line(&lb3, 5, EDIT, "Noise Decay", &el4);

	/* An edit field for the color scale factor. This is the
	 * number of sigma that the dark color of characters changes
	 * to beige, and half of that required to change to white.
	 */
	sprintf(eline5,"%.4f", pua43_colorscale); /* The field to be edited */
	sl=strlen(eline5);
	lb_create_el(&el5, 7, 135, &eline5[0], ED_INACTIVE, 0, sl);
	lb_line(&lb3, 2, EDIT, "Char Color Ratio", &el5);

	/* Define the Radio Button families:     */

	/* 14Ch/28Ch per msg family:     */
	if (pua14_28) rline=8; else rline=7;
	lb_create_rb(&rl1, 0, rline, 20);
	lb_line(&lb3, 7, RADIO, "14 Symbols/Msg", &rl1);
	lb_line(&lb3, 8, RADIO, "28 Symbols/Msg", &rl1);

	lb_radio_active(&rl1, 7, RADIO_ACTIVE); /* Start point in lb3 */
	lb_update(&lb3);
	goto r7;			/* Starting spot */

	/* Following is a loop for handling user keyboard input */

 e3:	edresult=lb_edit_line(&lb3, &el5);
	if (edresult==CUR_UP)   goto e16;
	else if (edresult==ENTER) goto lb3end;

 e5:	edresult=lb_edit_line(&lb3, &el4);
	if (edresult==CUR_UP)   goto e3;
	else if (edresult==ENTER) goto lb3end;

 r7:	/* Now give user opportunity to push button 7: */
	rbui=lb_radio_button(&lb3, 7, &rl1);
	/* Find out how exit was done and react:      */
	if     (rbui==CUR_UP)  goto e5;
	else if(rbui==ENTER)   goto lb3end;

 r8:	rbui=lb_radio_button(&lb3, 8, &rl1);
	if     (rbui==CUR_UP)  goto r7;
	else if(rbui==ENTER)   goto lb3end;

 e11:	edresult=lb_edit_line(&lb3, &el1);
	if (edresult==CUR_UP)   goto r8;
	else if (edresult==ENTER) goto lb3end;

 e14:	edresult=lb_edit_line(&lb3, &el2);
	if (edresult==CUR_UP)   goto e11;
	else if (edresult==ENTER) goto lb3end;

 e16:	edresult=lb_edit_line(&lb3, &el3);
	if (edresult==CUR_UP)   goto e14;
	else if (edresult==ENTER) goto lb3end;

 e18:	edresult=lb_edit_line(&lb3, &el6);
	if (edresult==CUR_UP)   goto e16;
	else if (edresult==ENTER) goto lb3end;
	else goto r7;

 lb3end: delay(200);
	/* As we exit, the variables must be changed to agree with the
	 * indicated states of the buttons and edit fields:
	 */
	pua14_28=1 & rl1.state[8];
	if (!pua14_28)
	    for (i=14; i<28; i++)
		{
		print_big(32, i, c_text_bold);
		print_super_row(32, i, c_text_bold);
		}
	strcpy(code_id, eline1);	/* Get any changes to code_id */
	strcpy(pua43_xstring, eline2);	/* First line */
	if (strlen(eline2) > 0)		/* If there is more */
	    strcat(pua43_xstring, eline3);
	if ((sl=strlen(pua43_xstring)) < 28)
	    {
	    for (i=sl; i<28; i++)
		pua43_xstring[i]=32;	/* Pad out with spaces */
	    pua43_xstring[28]=0;
	    }
	pua43_noisedecay=atof(eline4);
	pua43_colorscale=atof(eline5);

	nbdb=atof(eline6);
	if (nbdb < -99.9)
	    {
	    nbdb=-99.9;
	    lb_msg(170, 260, "WARNING",
		"  Noise Blank must be >-99.9",
		"      Value set to -99.9");
	    }
	if (nbdb > 99.9)
	    {
	    nbdb=99.9;
	    lb_msg(170, 260, "WARNING",
		"  Noise Blank max of 99.9 dB",
		"      Value set to 99.9");
	    }
	code_nb_ratio=upow10(0.1*nbdb);




	lb_close(&lb3, 1);
	}

/******************************************************
 * A modal dialog box for LTI Long Term Integration.
 */
void lterm_int_modalbox(void)
	{
	static struct edit_info   el1, el2, el3, el4, el5;
	static struct radio_info  rl2;
	static struct check_info  ck1;
	double nbdb, f;
	int bpush;
	char eline1[12], eline2[12], eline3[12], eline4[12], eline5[12];
	unsigned int edresult, rbui, ckui;

	lb_create(&lb4, "LTI Long Term Mode", 160, 175, 370, 440, 0, MODAL);
	lb_open(&lb4, 1);

	/* Fixed text in modal box:     */
	lb_line(&lb4, 21, BOLD, "\030 and \031 Move to Next Item", 0);
	lb_line(&lb4, 22, BOLD, "SPACE=Change Button/Box", 0);
	lb_line(&lb4, 23, BOLD, "TEXT: Backspace & Type", 0);
	lb_line(&lb4, 24, BOLD, "ENTER=Close Dialog", 0);

	/* Editable text areas:         */
	/* Note edit boxes take a line and the next one, ie el1 uses 4 and 5 */
	sprintf(eline1, "%d", (int)(0.5+lterm_int_flow));
	lb_create_el(&el1, 5, 143, &eline1[0], ED_INACTIVE, 0, strlen(eline1));
	lb_line(&lb4, 2, EDIT, "Display Fr Low, Hz", &el1);

	sprintf(eline3, "%d", (int)(0.5+lterm_int_center));
	lb_create_el(&el3, 5, 143, &eline3[0], ED_INACTIVE, 0, strlen(eline3));
	lb_line(&lb4, 4, EDIT, "Center Freq, Hz", &el3);

	sprintf(eline2, "%d", (int)(0.5+lterm_int_fhigh));
	lb_create_el(&el2, 5, 143, &eline2[0], ED_INACTIVE, 0, strlen(eline2));
	lb_line(&lb4, 6, EDIT, "Display Fr High,Hz", &el2);

	sprintf(eline5, "%.3d", (int)(lti_freq_range+0.5));
	lb_create_el(&el5, 5, 143, &eline5[0], ED_INACTIVE, 0, strlen(eline5));
	lb_line(&lb4, 8, EDIT, " Random Spread  Hz", &el5);

	if (code_nb_ratio<=0.0)  code_nb_ratio=0.10223294e-10;
	if (code_nb_ratio>=9.7723e9) code_nb_ratio=9.7723e9;
	sprintf(eline4, "%.2lf", 10.0*log10(code_nb_ratio));
	lb_create_el(&el4, 6, 143, &eline4[0], ED_INACTIVE, 0, strlen(eline4));
	lb_line(&lb4, 18, EDIT, "Noise Blank, dB", &el4);

	/* Define the max dB/div Radio Button family:     */
	if (min_lt_dd<=1.600001)
	    {
	    min_lt_dd=1.6;
	    bpush=13;
	    }
	else if (min_lt_dd<=3.200001)
	    {
	    min_lt_dd=3.2;
	    bpush=14;
	    }
	else if (min_lt_dd<=8.000001)
	    {
	    min_lt_dd=8.0;
	    bpush=15;
	    }
	else
	    {
	    min_lt_dd=16.0;
	    bpush=16;
	    }

	lb_create_rb(&rl2, 0, bpush, 20);	/* Note: bpush pushed */
	/* Each line is defined separately with lb_line(..):             */
	lb_line(&lb4, 13, RADIO, "LTerm Trace max 10 dB/", &rl2);
	lb_line(&lb4, 14, RADIO, "LTerm Trace max  5 dB/", &rl2);
	lb_line(&lb4, 15, RADIO, "LTerm Trace max  2 dB/", &rl2);
	lb_line(&lb4, 16, RADIO, "LTerm Trace max  1 dB/", &rl2);

	/* Record check box:     */
	if (eme_record)			/* Show box checked */
	    lb_create_check(&ck1, 0, TRUE, 20);
	else				/* Show box unchecked */
	    lb_create_check(&ck1, 0, FALSE, 20);
	lb_line(&lb4, 11, CHECK, "Record Data to Disk", &ck1);

	lb_update(&lb4);

 e2:	edresult=lb_edit_line(&lb4, &el1);
	if (edresult==CUR_UP)   goto e18;
	else if (edresult==ENTER) goto lb4end;

 e4:	edresult=lb_edit_line(&lb4, &el3);
	if (edresult==CUR_UP)   goto e2;
	else if (edresult==ENTER) goto lb4end;

 e6:	edresult=lb_edit_line(&lb4, &el2);
	if (edresult==CUR_UP)   goto e4;
	else if (edresult==ENTER) goto lb4end;

 e8:	edresult=lb_edit_line(&lb4, &el5);
	if (edresult==CUR_UP)   goto e6;
	else if (edresult==ENTER) goto lb4end;

 c11:	ckui=lb_check_mark(&lb4, &ck1);
	if     (ckui==CUR_UP)  goto e8;
	else if(ckui==ENTER)     goto lb4end;

 r13:	rbui=lb_radio_button(&lb4, 13, &rl2);
	if     (rbui==CUR_UP)  goto c11;
	else if(rbui==ENTER)   goto lb4end;

 r14:	rbui=lb_radio_button(&lb4, 14, &rl2);
	if     (rbui==CUR_UP)  goto r13;
	else if(rbui==ENTER)   goto lb4end;

 r15:	rbui=lb_radio_button(&lb4, 15, &rl2);
	if     (rbui==CUR_UP)  goto r14;
	else if(rbui==ENTER)   goto lb4end;

 r16:	rbui=lb_radio_button(&lb4, 16, &rl2);
	if     (rbui==CUR_UP)  goto r15;
	else if(rbui==ENTER)   goto lb4end;

 e18:	edresult=lb_edit_line(&lb4, &el4);
	if (edresult==CUR_UP)   goto r16;
	else if (edresult==CUR_DOWN) goto e2;
	else if (edresult==ENTER) goto lb4end;

 lb4end:/* As we exit, the variables must be changed to agree with the
	 * indicated states of the buttons and edit fields:
	 */

	f=atof(eline1);
	if (f<0.0)
	    {
	    f=0.0;
	    lb_msg(170, 260, "WARNING",
		" Frequencies must be positive",
		"    Spectrum Low set to 0");
	    }
	else if (f>2400.0)
	    {
	    f=1100.0;
	    lb_msg(170, 260, "WARNING",
		" Frequencies  must be < 2400",
		"   Spectrum Low set to 2300");
	    }
	lterm_int_flow=f;

	f=atof(eline2);
	if (f<0.0)
	    {
	    f=0.0;
	    lb_msg(170, 260, "WARNING",
		" Frequencies must be positive",
		"   Spectrum High set to 0");
	    }
	else if (f>2400.0)
	    {
	    f=2400.0;
	    lb_msg(170, 260, "WARNING",
		" Frequencies  must be < 2400",
		"  Spectrum High set to 2400");
	    }
	if (f < lterm_int_flow)
	    {
	    f=lterm_int_flow+100.0;
	    lb_msg(170, 260, "WARNING",
		"  Low must be less than High",
		"    Spectrum High adjusted");
	    }
	lterm_int_fhigh=f;

	f=atof(eline3);
	if (f<0.0)
	    {
	    f=100.0;
	    lb_msg(170, 260, "WARNING",
		" Frequencies must be positive",
		"  Spectrum Center set to 100");
	    }
	else if (f>1200.0)
	    {
	    f=1100.0;
	    lb_msg(170, 260, "WARNING",
		" Frequencies  must be < 2400",
		" Spectrum Center set to 2300");
	    }
	if (f<lterm_int_flow || f>lterm_int_fhigh)
	    {
	    f=0.5*(lterm_int_flow+lterm_int_fhigh);
	    lb_msg(170, 260, "WARNING",
		" Low < Center < High required",
		"   Spectrum Center adjusted");
	    }
	lterm_int_center=f;

	lti_freq_range=atof(eline5);

	if (ck1.state & CHECK_MARK)
	    eme_record=TRUE;
	else
	    eme_record=FALSE;

	if      ((1&rl2.state[13])==1) min_lt_dd= 1.6;
	else if ((1&rl2.state[14])==1) min_lt_dd= 3.2;
	else if ((1&rl2.state[15])==1) min_lt_dd= 8.0;
	else if ((1&rl2.state[16])==1) min_lt_dd=16.0;

	nbdb=atof(eline4);
	if (nbdb < -99.9)
	    {
	    nbdb=-99.9;
	    lb_msg(170, 260, "WARNING",
		"  Noise Blank must be >-99.9",
		"      Value set to -99.9");
	    }
	if (nbdb > 99.9)
	    {
	    nbdb=99.9;
	    lb_msg(170, 260, "WARNING",
		"  Noise Blank max of 99.9 dB",
		"      Value set to 99.9");
	    }
	code_nb_ratio=upow10(0.1*nbdb);

	set_bins();			/* Update variables */
	delay(500);
	lb_close(&lb4, 1);
	}
